import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from collections.abc import Iterable

from . import COLORS
from ..base import hex2rgb

RC_PARAMS = {
    # ***************************************************************************
    # * BACKENDS                                                                *
    # ***************************************************************************
    'interactive':                 False,

    # ***************************************************************************
    # * LINES                                                                   *
    # ***************************************************************************
    'lines.linewidth':             1.8,  # line width in points

    # ***************************************************************************
    # * PATHCES                                                                 *
    # ***************************************************************************

    # ***************************************************************************
    # * HATHCES                                                                 *
    # ***************************************************************************

    # ***************************************************************************
    # * BOXPLOT                                                                 *
    # ***************************************************************************

    # ***************************************************************************
    # * FONT                                                                    *
    # ***************************************************************************

    # ***************************************************************************
    # * TEXT                                                                    *
    # ***************************************************************************

    # ***************************************************************************
    # * LATEX                                                                   *
    # ***************************************************************************

    # ***************************************************************************
    # * AXES                                                                    *
    # ***************************************************************************
    'axes.titlesize':              'x-large',  # font size of the axes title
    'axes.formatter.use_mathtext': True,  # When True, use mathtext for scientific
    'axes.grid':                   True,  # display grid or not
    'axes.grid.which':             'both',  # grid lines at {major, minor, both} ticks
    'grid.color':                  COLORS.LIGHT_GREY,  # grid color

    'axes.labelsize':              'large',  # font size of the x and y labels

    'axes.prop_cycle':             matplotlib.cycler(
            color=[COLORS.DARK_BLUE, COLORS.ORANGE_B, COLORS.PURPLE_D, COLORS.GREEN, COLORS.GREY_BROWN,
                COLORS.GOLD, COLORS.MAROON_D, COLORS.BLUE_C, COLORS.BLACK, COLORS.PINK]),

    'axes.spines.left':            True,  # display axis spines
    'axes.spines.bottom':          True,
    'axes.spines.top':             False,
    'axes.spines.right':           False,

    # ***************************************************************************
    # * DATES                                                                   *
    # ***************************************************************************

    # ***************************************************************************
    # * TICKS                                                                   *
    # ***************************************************************************

    # ***************************************************************************
    # * GRIDS                                                                   *
    # ***************************************************************************

    # ***************************************************************************
    # * LEGEND                                                                  *
    # ***************************************************************************

    # ***************************************************************************
    # * FIGURE                                                                  *
    # ***************************************************************************
    'figure.autolayout':           True,  # When True, automatically adjust subplot
    'figure.figsize':              [8, 8],  # figure size in inches
    'figure.dpi':                  100,
    'figure.facecolor':            'none',  # figure face color
    'figure.edgecolor':            'none',  # figure edge color

    # ***************************************************************************
    # * IMAGES                                                                  *
    # ***************************************************************************

    # ***************************************************************************
    # * CONTOUR PLOTS                                                           *
    # ***************************************************************************

    # ***************************************************************************
    # * ERRORBAR PLOTS                                                          *
    # ***************************************************************************
    'errorbar.capsize':            3,

    # ***************************************************************************
    # * HISTOGRAM PLOTS                                                         *
    # ***************************************************************************

    # ***************************************************************************
    # * SCATTER PLOTS                                                           *
    # ***************************************************************************

    # ***************************************************************************
    # * AGG RENDERING                                                           *
    # ***************************************************************************

    # ***************************************************************************
    # * PATHS                                                                   *
    # ***************************************************************************

    # ***************************************************************************
    # * SAVING FIGURES                                                          *
    # ***************************************************************************
    'savefig.dpi':                 'figure',  # figure dots per inch or 'figure'
    'savefig.format':              'png',  # {png, ps, pdf, svg}
    'savefig.bbox':                'tight',  # {tight, standard}

    # ***************************************************************************
    # * INTERACTIVE KEYMAPS                                                     *
    # ***************************************************************************

    # ***************************************************************************
    # * ANIMATIONS                                                              *
    # ***************************************************************************
}

RC_PARAMS_DEFAULT = {
    ### MATPLOTLIBRC FORMAT

    # NOTE FOR END USERS: DO NOT EDIT THIS FILE!
    #
    # This is a sample Matplotlib configuration file - you can find a copy
    # of it on your system in site-packages/matplotlib/mpl-data/matplotlibrc
    # (relative to your Python installation location).
    #
    # You should find a copy of it on your system at
    # site-packages/matplotlib/mpl-data/matplotlibrc (relative to your Python
    # installation location).  DO NOT EDIT IT!
    #
    # If you wish to change your default style, copy this file to one of the
    # following locations:
    #     Unix/Linux:
    #         $HOME/.config/matplotlib/matplotlibrc OR
    #         $XDG_CONFIG_HOME/matplotlib/matplotlibrc (if $XDG_CONFIG_HOME is set)
    #     Other platforms:
    #         $HOME/.matplotlib/matplotlibrc
    # and edit that copy.
    #
    # See https://matplotlib.org/users/customizing.html#the-matplotlibrc-file
    # for more details on the paths which are checked for the configuration file.
    #
    # Blank lines, or lines starting with a comment symbol, are ignored, as are
    # trailing comments.  Other lines must have the format:
    #     key: val  # optional comment
    #
    # Formatting: Use PEP8-like style (as enforced in the rest of the codebase).
    # All lines start with an additional '#', so that removing all leading '#'s
    # yields a valid style file.
    #
    # Colors: for the color values below, you can either use
    #     - a Matplotlib color string, such as r, k, or b
    #     - an RGB tuple, such as (1.0, 0.5, 0.0)
    #     - a hex string, such as ff00ff
    #     - a scalar grayscale intensity such as 0.75
    #     - a legal html color name, e.g., red, blue, darkslategray
    #
    # Matplotlib configuration are currently divided into following parts:
    #     - BACKENDS
    #     - LINES
    #     - PATCHES
    #     - HATCHES
    #     - BOXPLOT
    #     - FONT
    #     - TEXT
    #     - LaTeX
    #     - AXES
    #     - DATES
    #     - TICKS
    #     - GRIDS
    #     - LEGEND
    #     - FIGURE
    #     - IMAGES
    #     - CONTOUR PLOTS
    #     - ERRORBAR PLOTS
    #     - HISTOGRAM PLOTS
    #     - SCATTER PLOTS
    #     - AGG RENDERING
    #     - PATHS
    #     - SAVING FIGURES
    #     - INTERACTIVE KEYMAPS
    #     - ANIMATION

    #### CONFIGURATION BEGINS HERE

    # ***************************************************************************
    # * BACKENDS                                                                *
    # ***************************************************************************
    # The default backend.  If you omit this parameter, the first working
    # backend from the following list is used:
    #     MacOSX Qt5Agg Gtk3Agg TkAgg WxAgg Agg
    # Other choices include:
    #     Qt5Cairo GTK3Cairo TkCairo WxCairo Cairo
    #     Qt4Agg Qt4Cairo Wx  # deprecated.
    #     PS PDF SVG Template
    # You can also deploy your own backend outside of Matplotlib by referring to
    # the module name (which must be in the PYTHONPATH) as 'module://my_backend'.
    'backend':                            'Agg',

    # The port to use for the web server in the WebAgg backend.
    'webagg.port':                        8988,

    # The address on which the WebAgg web server should be reachable
    'webagg.address':                     '127.0.0.1',

    # If webagg.port is unavailable, a number of other random ports will
    # be tried until one that is available is found.
    'webagg.port_retries':                50,

    # When True, open the web browser to the plot that is shown
    'webagg.open_in_browser':             True,

    # If you are running pyplot inside a GUI and your backend choice
    # conflicts, we will automatically try to find a compatible one for
    # you if backend_fallback is True
    'backend_fallback':                   True,

    'interactive':                        True,
    'toolbar':                            'toolbar2',  # {None, toolbar2, toolmanager}
    'timezone':                           'UTC',  # a pytz timezone string, e.g., US/Central or Europe/Paris

    # ***************************************************************************
    # * LINES                                                                   *
    # ***************************************************************************
    # See https://matplotlib.org/api/artist_api.html#module-matplotlib.lines
    # for more information on line properties.
    'lines.linewidth':                    1.5,  # line width in points
    'lines.linestyle':                    '-',  # solid line
    'lines.color':                        'C0',  # has no affect on plot(); see axes.prop_cycle
    'lines.marker':                       None,  # the default marker
    'lines.markerfacecolor':              'auto',  # the default marker face color
    'lines.markeredgecolor':              'auto',  # the default marker edge color
    'lines.markeredgewidth':              1.0,  # the line width around the marker symbol
    'lines.markersize':                   6,  # marker size, in points
    'lines.dash_joinstyle':               'round',  # {miter, round, bevel}
    'lines.dash_capstyle':                'butt',  # {butt, round, projecting}
    'lines.solid_joinstyle':              'round',  # {miter, round, bevel}
    'lines.solid_capstyle':               'projecting',  # {butt, round, projecting}
    'lines.antialiased':                  True,  # render lines in antialiased (no jaggies)

    # The three standard dash patterns.  These are scaled by the linewidth.
    'lines.dashed_pattern':               [3.7, 1.6],
    'lines.dashdot_pattern':              [6.4, 1.6, 1, 1.6],
    'lines.dotted_pattern':               [1, 1.65],
    'lines.scale_dashes':                 True,

    'markers.fillstyle':                  'full',  # {full, left, right, bottom, top, none}

    'pcolor.shading':                     'flat',
    'pcolormesh.snap':                    True,  # Whether to snap the mesh to pixel boundaries. This
    # is provided solely to allow old test images to remain
    # unchanged. Set to False to obtain the previous behavior.

    # ***************************************************************************
    # * PATCHES                                                                 *
    # ***************************************************************************
    # Patches are graphical objects that fill 2D space, like polygons or circles.
    # See https://matplotlib.org/api/artist_api.html#module-matplotlib.patches
    # for more information on patch properties.
    'patch.linewidth':                    1,  # edge width in points
    'patch.facecolor':                    'C0',
    'patch.edgecolor':                    'black',  # if forced, or patch is not filled
    'patch.force_edgecolor':              False,  # True to always use edgecolor
    'patch.antialiased':                  True,  # render patches in antialiased (no jaggies)

    # ***************************************************************************
    # * HATCHES                                                                 *
    # ***************************************************************************
    'hatch.color':                        'black',
    'hatch.linewidth':                    1.0,

    # ***************************************************************************
    # * BOXPLOT                                                                 *
    # ***************************************************************************
    'boxplot.notch':                      False,
    'boxplot.vertical':                   True,
    'boxplot.whiskers':                   1.5,
    'boxplot.bootstrap':                  None,
    'boxplot.patchartist':                False,
    'boxplot.showmeans':                  False,
    'boxplot.showcaps':                   True,
    'boxplot.showbox':                    True,
    'boxplot.showfliers':                 True,
    'boxplot.meanline':                   False,

    'boxplot.flierprops.color':           'black',
    'boxplot.flierprops.marker':          'o',
    'boxplot.flierprops.markerfacecolor': 'none',
    'boxplot.flierprops.markeredgecolor': 'black',
    'boxplot.flierprops.markeredgewidth': 1.0,
    'boxplot.flierprops.markersize':      6,
    'boxplot.flierprops.linestyle':       'none',
    'boxplot.flierprops.linewidth':       1.0,

    'boxplot.boxprops.color':             'black',
    'boxplot.boxprops.linewidth':         1.0,
    'boxplot.boxprops.linestyle':         '-',

    'boxplot.whiskerprops.color':         'black',
    'boxplot.whiskerprops.linewidth':     1.0,
    'boxplot.whiskerprops.linestyle':     '-',

    'boxplot.capprops.color':             'black',
    'boxplot.capprops.linewidth':         1.0,
    'boxplot.capprops.linestyle':         '-',

    'boxplot.medianprops.color':          'C1',
    'boxplot.medianprops.linewidth':      1.0,
    'boxplot.medianprops.linestyle':      '-',

    'boxplot.meanprops.color':            'C2',
    'boxplot.meanprops.marker':           '^',
    'boxplot.meanprops.markerfacecolor':  'C2',
    'boxplot.meanprops.markeredgecolor':  'C2',
    'boxplot.meanprops.markersize':       6,
    'boxplot.meanprops.linestyle':        '--',
    'boxplot.meanprops.linewidth':        1.0,

    # ***************************************************************************
    # * FONT                                                                    *
    # ***************************************************************************
    # The font properties used by `text.Text`.
    # See https://matplotlib.org/api/font_manager_api.html for more information
    # on font properties.  The 6 font properties used for font matching are
    # given below with their default values.
    #
    # The font.family property can take either a concrete font name (not supported
    # when rendering text with usetex), or one of the following five generic
    # values:
    #     - 'serif' (e.g., Times),
    #     - 'sans-serif' (e.g., Helvetica),
    #     - 'cursive' (e.g., Zapf-Chancery),
    #     - 'fantasy' (e.g., Western), and
    #     - 'monospace' (e.g., Courier).
    # Each of these values has a corresponding default list of font names
    # (font.serif, etc.); the first available font in the list is used.  Note that
    # for font.serif, font.sans-serif, and font.monospace, the first element of
    # the list (a DejaVu font) will always be used because DejaVu is shipped with
    # Matplotlib and is thus guaranteed to be available; the other entries are
    # left as examples of other possible values.
    #
    # The font.style property has three values: normal (or roman), italic
    # or oblique.  The oblique style will be used for italic, if it is not
    # present.
    #
    # The font.variant property has two values: normal or small-caps.  For
    # TrueType fonts, which are scalable fonts, small-caps is equivalent
    # to using a font size of 'smaller', or about 83%% of the current font
    # size.
    #
    # The font.weight property has effectively 13 values: normal, bold,
    # bolder, lighter, 100, 200, 300, ..., 900.  Normal is the same as
    # 400, and bold is 700.  bolder and lighter are relative values with
    # respect to the current weight.
    #
    # The font.stretch property has 11 values: ultra-condensed,
    # extra-condensed, condensed, semi-condensed, normal, semi-expanded,
    # expanded, extra-expanded, ultra-expanded, wider, and narrower.  This
    # property is not currently implemented.
    #
    # The font.size property is the default font size for text, given in points.
    # 10 pt is the standard value.
    #
    # Note that font.size controls default text sizes.  To configure
    # special text sizes tick labels, axes, labels, title, etc., see the rc
    # settings for axes and ticks.  Special text sizes can be defined
    # relative to font.size, using the following values: xx-small, x-small,
    # small, medium, large, x-large, xx-large, larger, or smaller

    'font.family':                        'sans-serif',
    'font.style':                         'normal',
    'font.variant':                       'normal',
    'font.weight':                        'normal',
    'font.stretch':                       'normal',
    'font.size':                          10.0,

    'font.serif':                         ['DejaVu Serif', 'Bitstream Vera Serif', 'Computer Modern Roman',
        'New Century Schoolbook', 'Century Schoolbook L', 'Utopia', 'ITC Bookman',
        'Bookman', 'Nimbus Roman No9 L', 'Times New Roman', 'Times', 'Palatino',
        'Charter', 'serif'],
    'font.sans-serif':                    ['DejaVu Sans', 'Bitstream Vera Sans', 'Computer Modern Sans Serif',
        'Lucida Grande', 'Verdana', 'Geneva', 'Lucid', 'Arial', 'Helvetica',
        'Avant Garde', 'sans-serif'],
    'font.cursive':                       ['Apple Chancery', 'Textile', 'Zapf Chancery', 'Sand', 'Script MT', 'Felipa',
        'Comic Neue', 'Comic Sans MS', 'cursive'],
    'font.fantasy':                       ['Chicago', 'Charcoal', 'Impact', 'Western', 'Humor Sans', 'xkcd', 'fantasy'],
    'font.monospace':                     ['DejaVu Sans Mono', 'Bitstream Vera Sans Mono', 'Computer Modern Typewriter',
        'Andale Mono', 'Nimbus Mono L', 'Courier New', 'Courier', 'Fixed',
        'Terminal', 'monospace'],

    # ***************************************************************************
    # * TEXT                                                                    *
    # ***************************************************************************
    # The text properties used by `text.Text`.
    # See https://matplotlib.org/api/artist_api.html#module-matplotlib.text
    # for more information on text properties
    'text.color':                         'black',

    # ***************************************************************************
    # * LaTeX                                                                   *
    # ***************************************************************************
    # For more information on LaTeX properties, see
    # https://matplotlib.org/tutorials/text/usetex.html
    'text.usetex':                        False,  # use latex for all text handling. The following fonts
    # are supported through the usual rc parameter settings:
    # new century schoolbook, bookman, times, palatino,
    # zapf chancery, charter, serif, sans-serif, helvetica,
    # avant garde, courier, monospace, computer modern roman,
    # computer modern sans serif, computer modern typewriter
    # If another font is desired which can loaded using the
    # LaTeX \usepackage command, please inquire at the
    # Matplotlib mailing list
    'text.latex.preamble':                '',  # IMPROPER USE OF THIS FEATURE WILL LEAD TO LATEX FAILURES
    # AND IS THEREFORE UNSUPPORTED. PLEASE DO NOT ASK FOR HELP
    # IF THIS FEATURE DOES NOT DO WHAT YOU EXPECT IT TO.
    # text.latex.preamble is a single line of LaTeX code that
    # will be passed on to the LaTeX system. It may contain
    # any code that is valid for the LaTeX "preamble", i.e.
    # between the "\documentclass" and "\begin{document}"
    # statements.
    # Note that it has to be put on a single line, which may
    # become quite long.
    # The following packages are always loaded with usetex, so
    # beware of package collisions: color, geometry, graphicx,
    # type1cm, textcomp.
    # Adobe Postscript (PSSNFS) font packages may also be
    # loaded, depending on your font settings.

    # FreeType hinting flag ("foo" corresponds to FT_LOAD_FOO); may be one of the
    # following (Proprietary Matplotlib-specific synonyms are given in parentheses,
    # but their use is discouraged):
    # - default: Use the font's native hinter if possible, else FreeType's auto-hinter.
    #            ("either" is a synonym).
    # - no_autohint: Use the font's native hinter if possible, else don't hint.
    #                ("native" is a synonym.)
    # - force_autohint: Use FreeType's auto-hinter.  ("auto" is a synonym.)
    # - no_hinting: Disable hinting.  ("none" is a synonym.)
    'text.hinting':                       'force_autohint',

    'text.hinting_factor':                8,  # Specifies the amount of softness for hinting in the
    # horizontal direction.  A value of 1 will hint to full
    # pixels.  A value of 2 will hint to half pixels etc.
    'text.kerning_factor':                0,  # Specifies the scaling factor for kerning values. This
    # is provided solely to allow old test images to remain
    # unchanged. Set to 6 to obtain previous behavior. Values
    # other than 0 or 6 have no defined meaning.
    'text.antialiased':                   True,  # If True (default), the text will be antialiased.
    # This only affects raster outputs.

    # The following settings allow you to select the fonts in math mode.
    'mathtext.fontset':                   'dejavusans',  # Should be 'dejavusans' (default),
    # 'dejavuserif', 'cm' (Computer Modern), 'stix',
    # 'stixsans' or 'custom' (unsupported, may go
    # away in the future)
    # "mathtext.fontset: custom" is defined by the mathtext.bf, .cal, .it, ...
    # settings which map a TeX font name to a fontconfig font pattern.  (These
    # settings are not used for other font sets.)
    'mathtext.bf':                        'sans:bold',
    'mathtext.cal':                       'cursive',
    'mathtext.it':                        'sans:italic',
    'mathtext.rm':                        'sans',
    'mathtext.sf':                        'sans',
    'mathtext.tt':                        'monospace',
    'mathtext.fallback':                  'cm',  # Select fallback font from ['cm' (Computer Modern), 'stix'
    # 'stixsans'] when a symbol can not be found in one of the
    # custom math fonts. Select 'None' to not perform fallback
    # and replace the missing character by a dummy symbol.
    'mathtext.default':                   'it',  # The default font to use for math.
    # Can be any of the LaTeX font names, including
    # the special name "regular" for the same font
    # used in regular text.

    # ***************************************************************************
    # * AXES                                                                    *
    # ***************************************************************************
    # Following are default face and edge COLORS, default tick sizes,
    # default font sizes for tick labels, and so on.  See
    # https://matplotlib.org/api/axes_api.html#module-matplotlib.axes
    'axes.facecolor':                     'white',  # axes background color
    'axes.edgecolor':                     'black',  # axes edge color
    'axes.linewidth':                     0.8,  # edge line width
    'axes.grid':                          True,  # display grid or not
    'axes.grid.axis':                     'both',  # which axis the grid should apply to
    'axes.grid.which':                    'both',  # grid lines at {major, minor, both} ticks
    'axes.titlelocation':                 'center',  # alignment of the title: {left, right, center}
    'axes.titlesize':                     'large',  # font size of the axes title
    'axes.titleweight':                   'normal',  # font weight of title
    'axes.titlecolor':                    'auto',  # color of the axes title, auto falls back to
    # text.color as default value
    'axes.titley':                        None,  # position title (axes relative units).  None implies auto
    'axes.titlepad':                      6.0,  # pad between axes and title in points
    'axes.labelsize':                     'large',  # font size of the x and y labels
    'axes.labelpad':                      4.0,  # space between label and axis
    'axes.labelweight':                   'normal',  # weight of the x and y labels
    'axes.labelcolor':                    'black',
    'axes.axisbelow':                     'line',  # draw axis gridlines and ticks:
    # - below patches (True)
    # - above patches but below lines ('line')
    # - above all (False)

    'axes.formatter.limits':              [-5, 6],  # use scientific notation if log10
    # of the axis range is smaller than the
    # first or larger than the second
    'axes.formatter.use_locale':          False,  # When True, format tick labels
    # according to the user's locale.
    # For example, use ',' as a decimal
    # separator in the fr_FR locale.
    'axes.formatter.use_mathtext':        True,  # When True, use mathtext for scientific
    # notation.
    'axes.formatter.min_exponent':        0,  # minimum exponent to format in scientific notation
    'axes.formatter.useoffset':           True,  # If True, the tick label formatter
    # will default to labeling ticks relative
    # to an offset when the data range is
    # small compared to the minimum absolute
    # value of the data.
    'axes.formatter.offset_threshold':    4,  # When useoffset is True, the offset
    # will be used when it can remove
    # at least this number of significant
    # digits from tick labels.

    'axes.spines.left':                   True,  # display axis spines
    'axes.spines.bottom':                 True,
    'axes.spines.top':                    True,
    'axes.spines.right':                  True,

    'axes.unicode_minus':                 True,  # use Unicode for the minus symbol rather than hyphen.  See
    # https://en.wikipedia.org/wiki/Plus_and_minus_signs#Character_codes
    'axes.prop_cycle':                    matplotlib.cycler('color',
            ['1f77b4', 'ff7f0e', '2ca02c', 'd62728', '9467bd', '8c564b', 'e377c2',
                '7f7f7f', 'bcbd22', '17becf']),
    # color cycle for plot lines as list of string color specs:
    # single letter, long name, or web-style hex
    # As opposed to all other parameters in this file, the color
    # values must be enclosed in quotes for this parameter,
    # e.g. '1f77b4', instead of 1f77b4.
    # See also https://matplotlib.org/tutorials/intermediate/color_cycle.html
    # for more details on prop_cycle usage.
    'axes.xmargin':                       .05,  # x margin.  See `axes.Axes.margins`
    'axes.ymargin':                       .05,  # y margin.  See `axes.Axes.margins`
    'axes.zmargin':                       .05,  # z margin.  See `axes.Axes.margins`
    'axes.autolimit_mode':                'data',  # If "data", use axes.xmargin and axes.ymargin as is.
    # If "round_numbers", after application of margins, axis
    # limits are further expanded to the nearest "round" number.
    'polaraxes.grid':                     True,  # display grid on polar axes
    'axes3d.grid':                        True,  # display grid on 3D axes

    # ***************************************************************************
    # * AXIS                                                                    *
    # ***************************************************************************
    'xaxis.labellocation':                'center',  # alignment of the xaxis label: {left, right, center}
    'yaxis.labellocation':                'center',  # alignment of the yaxis label: {bottom, top, center}

    # ***************************************************************************
    # * DATES                                                                   *
    # ***************************************************************************
    # These control the default format strings used in AutoDateFormatter.
    # Any valid format datetime format string can be used (see the python
    # `datetime` for details).  For example, by using:
    #     - '%%x' will use the locale date representation
    #     - '%%X' will use the locale time representation
    #     - '%%c' will use the full locale datetime representation
    # These values map to the scales:
    #     {'year': 365, 'month': 30, 'day': 1, 'hour': 1/24, 'minute': 1 / (24 * 60)}

    'date.autoformatter.year':            '%Y',
    'date.autoformatter.month':           '%Y-%m',
    'date.autoformatter.day':             '%Y-%m-%d',
    'date.autoformatter.hour':            '%m-%d %H',
    'date.autoformatter.minute':          '%d %H:%M',
    'date.autoformatter.second':          '%H:%M:%S',
    'date.autoformatter.microsecond':     '%M:%S.%f',
    # The reference date for Matplotlib's internal date representation
    # See https://matplotlib.org/examples/ticks_and_spines/date_precision_and_epochs.py
    'date.epoch':                         '1970-01-01T00:00:00',
    # 'auto', 'concise':
    'date.converter':                     'auto',
    # For auto converter whether to use interval_multiples:
    'date.interval_multiples':            True,

    # ***************************************************************************
    # * TICKS                                                                   *
    # ***************************************************************************
    # See https://matplotlib.org/api/axis_api.html#matplotlib.axis.Tick
    'xtick.top':                          False,  # draw ticks on the top side
    'xtick.bottom':                       True,  # draw ticks on the bottom side
    'xtick.labeltop':                     False,  # draw label on the top
    'xtick.labelbottom':                  True,  # draw label on the bottom
    'xtick.major.size':                   3.5,  # major tick size in points
    'xtick.minor.size':                   2,  # minor tick size in points
    'xtick.major.width':                  0.8,  # major tick width in points
    'xtick.minor.width':                  0.6,  # minor tick width in points
    'xtick.major.pad':                    3.5,  # distance to major tick label in points
    'xtick.minor.pad':                    3.4,  # distance to the minor tick label in points
    'xtick.color':                        'black',  # color of the ticks
    'xtick.labelcolor':                   'inherit',  # color of the tick labels or inherit from xtick.color
    'xtick.labelsize':                    'medium',  # font size of the tick labels
    'xtick.direction':                    'out',  # direction: {in, out, inout}
    'xtick.minor.visible':                False,  # visibility of minor ticks on x-axis
    'xtick.major.top':                    True,  # draw x axis top major ticks
    'xtick.major.bottom':                 True,  # draw x axis bottom major ticks
    'xtick.minor.top':                    True,  # draw x axis top minor ticks
    'xtick.minor.bottom':                 True,  # draw x axis bottom minor ticks
    'xtick.alignment':                    'center',  # alignment of xticks

    'ytick.left':                         True,  # draw ticks on the left side
    'ytick.right':                        False,  # draw ticks on the right side
    'ytick.labelleft':                    True,  # draw tick labels on the left side
    'ytick.labelright':                   False,  # draw tick labels on the right side
    'ytick.major.size':                   3.5,  # major tick size in points
    'ytick.minor.size':                   2,  # minor tick size in points
    'ytick.major.width':                  0.8,  # major tick width in points
    'ytick.minor.width':                  0.6,  # minor tick width in points
    'ytick.major.pad':                    3.5,  # distance to major tick label in points
    'ytick.minor.pad':                    3.4,  # distance to the minor tick label in points
    'ytick.color':                        'black',  # color of the ticks
    'ytick.labelcolor':                   'inherit',  # color of the tick labels or inherit from ytick.color
    'ytick.labelsize':                    'medium',  # font size of the tick labels
    'ytick.direction':                    'out',  # direction: {in, out, inout}
    'ytick.minor.visible':                False,  # visibility of minor ticks on y-axis
    'ytick.major.left':                   True,  # draw y axis left major ticks
    'ytick.major.right':                  True,  # draw y axis right major ticks
    'ytick.minor.left':                   True,  # draw y axis left minor ticks
    'ytick.minor.right':                  True,  # draw y axis right minor ticks
    'ytick.alignment':                    'center_baseline',  # alignment of yticks

    # ***************************************************************************
    # * GRIDS                                                                   *
    # ***************************************************************************
    'grid.color':                         COLORS.DARK_GREY,  # grid color
    'grid.linestyle':                     '-',  # solid
    'grid.linewidth':                     0.8,  # in points
    'grid.alpha':                         1.0,  # transparency, between 0.0 and 1.0

    # ***************************************************************************
    # * LEGEND                                                                  *
    # ***************************************************************************
    'legend.loc':                         'best',
    'legend.frameon':                     True,  # if True, draw the legend on a background patch
    'legend.framealpha':                  0.8,  # legend patch transparency
    'legend.facecolor':                   'inherit',  # inherit from axes.facecolor; or color spec
    'legend.edgecolor':                   '0.8',  # background patch boundary color
    'legend.fancybox':                    True,  # if True, use a rounded box for the
    # legend background, else a rectangle
    'legend.shadow':                      False,  # if True, give background a shadow effect
    'legend.numpoints':                   1,  # the number of marker points in the legend line
    'legend.scatterpoints':               1,  # number of scatter points
    'legend.markerscale':                 1.0,  # the relative size of legend markers vs. original
    'legend.fontsize':                    'medium',
    'legend.title_fontsize':              None,  # None sets to the same as the default axes.

    # Dimensions as fraction of font size:
    'legend.borderpad':                   0.4,  # border whitespace
    'legend.labelspacing':                0.5,  # the vertical space between the legend entries
    'legend.handlelength':                2.0,  # the length of the legend lines
    'legend.handleheight':                0.7,  # the height of the legend handle
    'legend.handletextpad':               0.8,  # the space between the legend line and legend text
    'legend.borderaxespad':               0.5,  # the border between the axes and legend edge
    'legend.columnspacing':               2.0,  # column separation

    # ***************************************************************************
    # * FIGURE                                                                  *
    # ***************************************************************************
    # See https://matplotlib.org/api/figure_api.html#matplotlib.figure.Figure
    'figure.titlesize':                   'large',  # size of the figure title (``Figure.suptitle()``)
    'figure.titleweight':                 'normal',  # weight of the figure title
    'figure.figsize':                     [10, 10],  # figure size in inches
    'figure.dpi':                         100,  # figure dots per inch
    'figure.facecolor':                   'white',  # figure face color
    'figure.edgecolor':                   'white',  # figure edge color
    'figure.frameon':                     True,  # enable figure frame
    'figure.max_open_warning':            20,  # The maximum number of figures to open through
    # the pyplot interface before emitting a warning.
    # If less than one this feature is disabled.
    'figure.raise_window':                True,  # Raise the GUI window to front when show() is called.

    # The figure subplot parameters.  All dimensions are a fraction of the figure width and height.
    'figure.subplot.left':                0.125,  # the left side of the subplots of the figure
    'figure.subplot.right':               0.9,  # the right side of the subplots of the figure
    'figure.subplot.bottom':              0.11,  # the bottom of the subplots of the figure
    'figure.subplot.top':                 0.88,  # the top of the subplots of the figure
    'figure.subplot.wspace':              0.2,  # the amount of width reserved for space between subplots,
    # expressed as a fraction of the average axis width
    'figure.subplot.hspace':              0.2,  # the amount of height reserved for space between subplots,
    # expressed as a fraction of the average axis height

    # Figure layout
    'figure.autolayout':                  True,  # When True, automatically adjust subplot
    # parameters to make the plot fit the figure
    # using `tight_layout`
    'figure.constrained_layout.use':      False,  # When True, automatically make plot
    # elements fit on the figure. (Not
    # compatible with `autolayout`, above).
    'figure.constrained_layout.h_pad':    0.04167,  # Padding around axes objects. Float representing
    'figure.constrained_layout.w_pad':    0.04167,  # inches. Default is 3/72 inches (3 points)
    'figure.constrained_layout.hspace':   0.02,  # Space between subplot groups. Float representing
    'figure.constrained_layout.wspace':   0.02,  # a fraction of the subplot widths being separated.

    # ***************************************************************************
    # * IMAGES                                                                  *
    # ***************************************************************************
    'image.aspect':                       'equal',  # {equal, auto} or a number
    'image.interpolation':                'antialiased',  # see help(imshow) for options
    'image.cmap':                         'viridis',  # A colormap name, gray etc...
    'image.lut':                          256,  # the size of the colormap lookup table
    'image.origin':                       'upper',  # {lower, upper}
    'image.resample':                     True,
    'image.composite_image':              True,  # When True, all the images on a set of axes are
    # combined into a single composite image before
    # saving a figure as a vector graphics file,
    # such as a PDF.

    # ***************************************************************************
    # * CONTOUR PLOTS                                                           *
    # ***************************************************************************
    'contour.negative_linestyle':         'dashed',  # string or on-off ink sequence
    'contour.corner_mask':                True,  # {True, False, legacy}
    'contour.linewidth':                  None,  # {float, None} Size of the contour line
    # widths. If set to None, it falls back to
    # `line.linewidth`.

    # ***************************************************************************
    # * ERRORBAR PLOTS                                                          *
    # ***************************************************************************
    'errorbar.capsize':                   0,  # length of end cap on error bars in pixels

    # ***************************************************************************
    # * HISTOGRAM PLOTS                                                         *
    # ***************************************************************************
    'hist.bins':                          10,  # The default number of histogram bins or 'auto'.

    # ***************************************************************************
    # * SCATTER PLOTS                                                           *
    # ***************************************************************************
    'scatter.marker':                     'o',  # The default marker type for scatter plots.
    'scatter.edgecolors':                 'face',  # The default edge COLORS for scatter plots.

    # ***************************************************************************
    # * AGG RENDERING                                                           *
    # ***************************************************************************
    # Warning: experimental, 2008/10/10
    'agg.path.chunksize':                 0,  # 0 to disable; values in the range
    # 10000 to 100000 can improve speed slightly
    # and prevent an Agg rendering failure
    # when plotting very large data sets,
    # especially if they are very gappy.
    # It may cause minor artifacts, though.
    # A value of 20000 is probably a good
    # starting point.

    # ***************************************************************************
    # * PATHS                                                                   *
    # ***************************************************************************
    'path.simplify':                      True,  # When True, simplify paths by removing "invisible"
    # points to reduce file size and increase rendering
    # speed
    'path.simplify_threshold':            0.111111111111,  # The threshold of similarity below
    # which vertices will be removed in
    # the simplification process.
    'path.snap':                          True,  # When True, rectilinear axis-aligned paths will be snapped
    # to the nearest pixel when certain criteria are met.
    # When False, paths will never be snapped.
    'path.sketch':                        None,  # May be None, or a 3-tuple of the form:
    # (scale, length, randomness).
    #     - *scale* is the amplitude of the wiggle
    #         perpendicular to the line (in pixels).
    #     - *length* is the length of the wiggle along the
    #         line (in pixels).
    #     - *randomness* is the factor by which the length is
    #         randomly scaled.
    # 'path.effects':                       '',

    # ***************************************************************************
    # * SAVING FIGURES                                                          *
    # ***************************************************************************
    # The default savefig parameters can be different from the display parameters
    # e.g., you may want a higher resolution, or to make the figure
    # background white
    'savefig.dpi':                        'figure',  # figure dots per inch or 'figure'
    'savefig.facecolor':                  'auto',  # figure face color when saving
    'savefig.edgecolor':                  'auto',  # figure edge color when saving
    'savefig.format':                     'png',  # {png, ps, pdf, svg}
    'savefig.bbox':                       'tight',  # {tight, standard}
    # 'tight' is incompatible with pipe-based animation
    # backends (e.g. 'ffmpeg') but will work with those
    # based on temporary files (e.g. 'ffmpeg_file')
    'savefig.pad_inches':                 0.1,  # Padding to be used when bbox is set to 'tight'
    'savefig.directory':                  '~',  # default directory in savefig dialog box,
    # leave empty to always use current working directory
    'savefig.transparent':                False,  # setting that controls whether figures are saved with a
    # transparent background by default
    'savefig.orientation':                'portrait',  # Orientation of saved figure

    ## tk backend params
    'tk.window_focus':                    False,  # Maintain shell focus for TkAgg

    ## ps backend params
    'ps.papersize':                       'letter',  # {auto, letter, legal, ledger, A0-A10, B0-B10}
    'ps.useafm':                          False,  # use of AFM fonts, results in small files
    'ps.usedistiller':                    False,  # {ghostscript, xpdf, None}
    # Experimental: may produce smaller files.
    # xpdf intended for production of publication quality files,
    # but requires ghostscript, xpdf and ps2eps
    'ps.distiller.res':                   6000,  # dpi
    'ps.fonttype':                        3,  # Output Type 3 (Type3) or Type 42 (TrueType)

    ## PDF backend params
    'pdf.compression':                    6,  # integer from 0 to 9
    # 0 disables compression (good for debugging)
    'pdf.fonttype':                       3,  # Output Type 3 (Type3) or Type 42 (TrueType)
    'pdf.use14corefonts':                 False,
    'pdf.inheritcolor':                   False,

    ## SVG backend params
    'svg.image_inline':                   True,  # Write raster image data directly into the SVG file
    'svg.fonttype':                       'path',  # How to handle SVG fonts:
    # path: Embed characters as paths -- supported
    #       by most SVG renderers
    # None: Assume fonts are installed on the
    #       machine where the SVG will be viewed.
    'svg.hashsalt':                       None,  # If not None, use this string as hash salt instead of uuid4

    ## pgf parameter
    # See https://matplotlib.org/tutorials/text/pgf.html for more information.
    'pgf.rcfonts':                        True,
    # 'pgf.preamble':                       '',  # See text.latex.preamble for documentation
    'pgf.texsystem':                      'xelatex',

    ## docstring params
    'docstring.hardcopy':                 False,  # set this when you want to generate hardcopy docstring

    # ***************************************************************************
    # * INTERACTIVE KEYMAPS                                                     *
    # ***************************************************************************
    # Event keys to interact with figures/plots via keyboard.
    # See https://matplotlib.org/users/navigation_toolbar.html for more details on
    # interactive navigation.  Customize these settings according to your needs.
    # Leave the field(s) empty if you don't need a key-map. (i.e., fullscreen : '')
    'keymap.fullscreen':                  ['f', 'ctrl+f'],  # toggling
    'keymap.home':                        ['h', 'r', 'home'],  # home or reset mnemonic
    'keymap.back':                        ['left', 'c', 'backspace', 'MouseButton.BACK'],  # forward / backward keys
    'keymap.forward':                     ['right', 'v', 'MouseButton.FORWARD'],  # for quick navigation
    'keymap.pan':                         'p',  # pan mnemonic
    'keymap.zoom':                        "o",  # zoom mnemonic
    'keymap.save':                        ['s', 'ctrl+s'],  # saving current figure
    'keymap.help':                        'f1',  # display help about active tools
    'keymap.quit':                        ['ctrl+w', 'cmd+w', 'q'],  # close the current figure
    # 'keymap.quit_all':                    '',  # close all figures
    'keymap.grid':                        'g',  # switching on/off major grids in current axes
    'keymap.grid_minor':                  'G',  # switching on/off minor grids in current axes
    'keymap.yscale':                      'l',  # toggle scaling of y-axes ('log'/'linear')
    'keymap.xscale':                      ['k', 'L'],  # toggle scaling of x-axes ('log'/'linear')
    'keymap.copy':                        ['ctrl+c', 'cmd+c'],  # Copy figure to clipboard

    # ***************************************************************************
    # * ANIMATION                                                               *
    # ***************************************************************************
    'animation.html':                     'none',  # How to display the animation as HTML in
    # the IPython notebook:
    #     - 'html5' uses HTML5 video tag
    #     - 'jshtml' creates a JavaScript animation
    'animation.writer':                   'ffmpeg',  # MovieWriter 'backend' to use
    'animation.codec':                    'h264',  # Codec to use for writing movie
    'animation.bitrate':                  -1,  # Controls size/quality trade-off for movie.
    # -1 implies let utility auto-determine
    'animation.frame_format':             'png',  # Controls frame format used by temp files
    'animation.ffmpeg_path':              'ffmpeg',  # Path to ffmpeg binary. Without full path
    # $PATH is searched
    # 'animation.ffmpeg_args':              '',  # Additional arguments to pass to ffmpeg
    'animation.convert_path':             'convert',  # Path to ImageMagick's convert binary.
    # On Windows use the full path since convert
    # is also the name of a system tool.
    # 'animation.convert_args':             '',  # Additional arguments to pass to convert
    'animation.embed_limit':              20.0,  # Limit, in MB, of size of base64 encoded
    # animation in HTML (i.e. IPython notebook)

}


def get_style_rcparams(style: str):
    return plt.style.core._base_library[style]


STYLES = plt.style.core._base_library | {
    "default":                matplotlib.rcParamsDefault,

    "liron-utils-default":    RC_PARAMS,

    "liron-utils-article":    {  # get_style_rcparams("seaborn-v0_8-talk")
        # BACKENDS
        # LINES
        'lines.linewidth':       2.275,  # line width in points
        'lines.markeredgewidth': 0.0,  # the line width around the marker symbol
        'lines.markersize':      9.1,  # marker size, in points
        # PATCHES
        'patch.linewidth':       0.39,  # edge width in points.
        # HATCHES
        # BOXPLOT
        # FONT
        # TEXT
        # LaTeX
        # AXES
        'axes.labelsize':        14.3,  # font size of the x and y labels
        'axes.titlesize':        15.6,  # font size of the axes title
        'axes.grid':             False,  # display grid or not
        'axes.spines.left':      True,  # display axis spines
        'axes.spines.bottom':    True,
        'axes.spines.top':       True,
        'axes.spines.right':     True,
        # DATES
        # TICKS
        'xtick.labelsize':       13.0,  # font size of the tick labels
        'xtick.major.pad':       9.1,  # distance to major tick label in points
        'xtick.major.width':     1.3,  # major tick width in points
        'xtick.minor.width':     0.65,  # minor tick width in points
        'ytick.labelsize':       13.0,  # font size of the tick labels
        'ytick.major.pad':       9.1,  # distance to major tick label in points
        'ytick.major.width':     1.3,  # major tick width in points
        'ytick.minor.width':     0.65,  # minor tick width in points
        'xtick.top':             True,  # draw ticks on the top side
        'xtick.direction':       'in',  # direction: {in, out, inout}
        'ytick.right':           True,  # draw ticks on the right side
        'ytick.direction':       'in',  # direction: {in, out, inout}
        # GRIDS
        'grid.linewidth':        1.3,  # in points
        # LEGEND
        'legend.fontsize':       13.0,
        # FIGURE
        'figure.figsize':        [10.4, 7.15],  # figure size in inches
        'figure.titlesize':      15.6,  # size of the figure title (``Figure.suptitle()``)
        # IMAGES
        # CONTOUR PLOTS
        # ERRORBAR PLOTS
        # HISTOGRAM PLOTS
        # SCATTER PLOTS
        # AGG RENDERING
        # PATHS
        # SAVING FIGURES
        'savefig.pad_inches':    0,  # Padding to be used when bbox is set to 'tight'
        # INTERACTIVE KEYMAPS
        # ANIMATION
    },

    "liron-utils-text-color": lambda color: {
        # BACKENDS
        # LINES
        # PATCHES
        # HATCHES
        # BOXPLOT
        # FONT
        # TEXT
        # LaTeX
        # AXES
        'axes.titlecolor': color,
        'axes.labelcolor': color,
        # DATES
        # TICKS
        'xtick.color':     color,
        'ytick.color':     color,
        # GRIDS
        # LEGEND
        # FIGURE
        # IMAGES
        # CONTOUR PLOTS
        # ERRORBAR PLOTS
        # HISTOGRAM PLOTS
        # SCATTER PLOTS
        # AGG RENDERING
        # PATHS
        # SAVING FIGURES
        # INTERACTIVE KEYMAPS
        # ANIMATION

    },

    "liron-utils-notebook":   {
        # BACKENDS (need to run '%matplotlib Notebook')
        'backend':          'nbAgg',
        'interactive':      True,
        # LINES
        # PATCHES
        # HATCHES
        # BOXPLOT
        # FONT
        # TEXT
        # LaTeX
        # AXES
        'axes.titlesize':   'small',  # font size of the axes title
        'axes.labelsize':   'small',  # font size of the x and y labels
        # DATES
        # TICKS
        'xtick.labelsize':  'x-small',  # font size of the tick labels
        'ytick.labelsize':  'x-small',  # font size of the tick labels
        # GRIDS
        # LEGEND
        'legend.fontsize':  'x-small',
        # FIGURE
        'figure.titlesize': 'small',  # size of the figure title (``Figure.suptitle()``)
        'figure.figsize':   [6.4, 4.8],  # figure size in inches
        # IMAGES
        # CONTOUR PLOTS
        # ERRORBAR PLOTS
        # HISTOGRAM PLOTS
        # SCATTER PLOTS
        # AGG RENDERING
        # PATHS
        # SAVING FIGURES
        # INTERACTIVE KEYMAPS
        # ANIMATION

    }
}


def get_color_cycler(colors: (str, Iterable) = "cycler", rgb: bool = False, plot: (bool, str) = False):
    """
    Show all of a cycler colors in a graph

    Parameters
    ----------
    colors :        str or array_like, optional
                    - "all" - return all colors given in COLORS.py
                    - "cycler" - return only cycler colors given in RC_PARAMS["axes.prop_cycle"]
                    - "<color>" - return all colors in COLORS.py that start with <color> (e.g. "blue")
    rgb :           bool, optional
                    If True, return RGB tuples instead of hex strings
    plot :          bool, optional
                    Plot the colors

    Returns
    -------
    list of colors (hex or RGB)
    """

    if type(colors) is str:
        if colors == "cycler":
            cycler = RC_PARAMS["axes.prop_cycle"]
            colors = [cycler._left[i]['color'] for i in range(len(cycler._left))]
        elif colors == "all":
            all_str = COLORS.__all__
            colors = [getattr(COLORS, all_str[i]) for i in range(len(all_str))]
        else:
            color_name = colors
            colors = [getattr(COLORS, c) for c in COLORS.__all__ if color_name.lower() in c.lower()]
            if len(colors) == 0:
                raise ValueError(f"No colors found that match '{color_name}' in COLORS.py")

    elif isinstance(colors, Iterable):
        pass

    else:
        raise ValueError("Invalid type for 'colors'.")

    # plot colors
    if plot is not False:
        fig, ax = plt.subplots()
        if plot is True or plot in ["color", "colors"]:
            ax.axes.yaxis.set_visible(False)
            for i in range(len(colors)):
                h = ax.bar(i + 1, 1, color=colors[i])
                ax.bar_label(h, labels=[colors[i]], label_type="center", rotation=-90)
            ax.grid(None)

        elif plot == "sine":
            x = np.linspace(0, 1, 1001)
            y = lambda f: np.sin(2 * np.pi * f * x)
            for i in range(len(colors)):
                ax.plot(x, y((i + 1) / 4), label=colors[i])
                ax.legend()

        fig.show()

    if rgb:
        colors = hex2rgb(colors)

    return colors


def set_color_cycler(colors: (str, Iterable) = "cycler", plot: (bool, str) = False):
    """
    Set the color cycler for MatPlotLib.

    Parameters
    ----------
    colors :        str or array_like, optional
                    - "all" - set all colors given in COLORS.py
                    - "cycler" - set only cycler colors given in RC_PARAMS["axes.prop_cycle"]
    plot :          bool, optional
                    Plot the colors

    Returns
    -------
    list of colors
    """

    cycler_colors = get_color_cycler(colors=colors, plot=plot)
    matplotlib.rcParams['axes.prop_cycle'] = matplotlib.cycler(color=cycler_colors)

    return cycler_colors


def update_rcParams(new_params: (str, dict, callable) = "liron-utils-default", *args, **kwargs):
    """
    Update the default MatPlotLib rcParams.

    Parameters
    ----------
    new_params :    str or dict
                    - str: either one of plt.style.available or one of STYLES
                    - dict: dictionary of rcParams

    Returns
    -------

    """

    if type(new_params) is str:
        if new_params.replace("_", "-") in STYLES:
            new_params = STYLES[new_params.replace("_", "-")]
            if callable(new_params):
                new_params = new_params(*args, **kwargs)
            matplotlib.rcParams.update(new_params)
        else:  # new_rcParams in plt.style.available
            plt.style.use(new_params)
            new_params = get_style_rcparams(new_params)

    else:  # type(new_rcParams) is dict
        matplotlib.rcParams.update(new_params)

    return new_params



import time
from tqdm import tqdm
from liron_utils.pure_python import parallel


def foo(i, x, y):
	time.sleep(0.1)
	return i


def no_parallel(func, iterable, **kwargs) -> list:
	"""Run a function on an iterable without parallelization."""
	return [func(i, **kwargs) for i in tqdm(iterable)]


if __name__ == '__main__':
	for par_func in [parallel.parallel_map, parallel.parallel_threading]:
		t0 = time.time()
		out = par_func(func=foo, iterable=range(500), x=1, y=2,
				tqdm_kw=dict(desc=par_func.__name__, postfix=lambda i: dict(iter=i)))
		print(out[:5])
		print(f"{par_func.__name__} time: {time.time() - t0:.3f} sec")
		time.sleep(0.1)

	"""
	Results:
	--------
	len(iterable)	                    | 100   | 1000  | 10000 |
	parallel_map                        | 4.0   | 10.2  | 72.7  | [sec]
	parallel_threading (10 threads)     | 1.1   | 11.1  | 111.2 | [sec]
	"""
	pass



from scipy.signal import windows

from . import COLORS
from ...pure_python import MetaDict


class _FuncDefaultKwargs(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class DefaultKwargs(metaclass=MetaDict):
    # Figure
    FIG_KW = _FuncDefaultKwargs(

    )

    # Axes
    SET_PROPS_KW = _FuncDefaultKwargs(
            sup_title=None,  # str
            ax_title=None,  # list(str)
            axis=None,  # list(bool)
            spines=None,  # list(bool)
            ticks=None,  # list(bool/list)
            tick_labels=None,  # list(bool/list)
            labels=None,  # list(list(str))
            limits=None,  # list(list(float))
            view=None,  # list(list(float))
            grid=None,  # list(bool)
            legend=True,  # list(bool/list(str))
            legend_loc=None,  # list(str)
            colorbar=False,  # list(bool/list)
            xy_lines=True,  # list(bool)
            face_color=None,  # list(color)
            show_fig=True,  # bool
            open_dir=False,  # bool
            close_fig=False  # bool
    )

    XY_LINES_KW = _FuncDefaultKwargs(
            color=COLORS.DARK_GREY,
            linewidth=2
    )

    # 2D Plotting
    PLOT_KW = _FuncDefaultKwargs(

    )

    ERRORBAR_KW = _FuncDefaultKwargs(
            linestyle="none",
            marker=".",
            markersize=10,
            ecolor=COLORS.RED_E,
            elinewidth=1.4
    )

    FILL_BETWEEN_KW = _FuncDefaultKwargs(
            linestyle="-",
            color=COLORS.LIGHT_GRAY,
            alpha=0.4
    )

    SPECGRAM_KW = _FuncDefaultKwargs(
            NFFT=4096,
            window=windows.blackmanharris(4096),
            noverlap=int(0.85 * 4096),
            pad_to=4096 + int(1 * 4096),
            cmap='inferno'
    )

    # 3D Plotting
    PLOT_SURFACE_KW = _FuncDefaultKwargs(
            cmap='viridis',
    )


def update_kwargs(key=None, **kwargs):
    """
    Update the default kwargs with the new ones.

    Args:
    	key (str):  Key to KWARGS to update.
    	**kwargs:   New kwargs to update or merge.

    Returns:
    	dict: Updated KWARGS.
    """

    if key:
        DefaultKwargs[key.upper()].update(**kwargs)
    else:
        for key in kwargs.keys():
            DefaultKwargs[key.upper()].update(**kwargs[key])
    return DefaultKwargs


def merge_kwargs(**kwargs):
    """
    Merge between empty/partially filled kwargs to the default ones, giving priority to the new settings.

    Args:
    	**kwargs:

    Returns:
    	KWARGS " kwargs (take KWARGS and overwrite it with kwargs where needed)
    """

    for key in kwargs.keys():
        if kwargs[key] is None:
            kwargs[key] = dict()

        kwargs[key] = DefaultKwargs[key.upper()] | kwargs[key]

    return kwargs



from .base import *
from .dicts import *
from .docstring import *
from .imports import *
from .logs import *
from .os import *
from .prints import *
from .parallel import *
# from .pip import *
from .progress_bar import *

__all__ = [s for s in dir() if not s.startswith('_')]



from .base import *
from .axes import *
from .plotting import *
from .utils import *
from ..pure_python import is_notebook

__all__ = [s for s in dir() if not s.startswith('_')]

update_rcParams()  # Change default MatPlotLib parameters (e.g, figure size, label size, grid, colors, etc.)

if is_notebook():
    update_rcParams("liron-utils-notebook")

# TODO:
#   - matplotlib.animation.FuncAnimation
#   - Transfer my default kwargs to merge with mpl.rcParams



# from import_tracker import track_module
# from pprint import pprint
#
# x = track_module(module_name="liron_utils.graphics", track_import_stack=True, show_optional=True)
# pprint(x)

import numpy as np
from scipy.signal import windows
from liron_utils import graphics as gr

gr.update_rcParams("liron-utils-text_color", "white")

Ax = gr.Axes(shape=(2, 1))
gr.Axes(axs=Ax.axs[0, 0]).plot_fft(windows.hamming(1000), one_sided=False, dB=True, which="power")
Ax.axs[0, 0].set_ylim([-100, 0])
gr.Axes(axs=Ax.axs[1, 0]).plot_fft(windows.hamming(1000), one_sided=False, dB=True, which="phase")
gr.set_props()

# import scipy.optimize
# import matplotlib.pyplot as plt
#
# N = 101
# x = np.linspace(0, 10, N)
# yerr = 5 * np.random.randn(N)
# y = 2 * x ** 2 + 4 * x + 5 + yerr
#
#
# def fit_fcn(x, a, b, c):
#     return a * x ** 2 + b * x + c
#
# (p_opt, p_cov) = scipy.optimize.curve_fit(fit_fcn, x, y)
#
# fig, ax = gr.new_figure()
# gr.draw_xy_lines(ax)
# ax.plot(range(5))
# # gr.plot_data_and_curve_fit(x, y, fit_fcn, yerr=yerr, p_opt=p_opt, p_cov=p_cov)
# # ax.show_fig()
# fig.show()
# pass

# Ax = gr.AxesLiron(2, 1)
# t = np.linspace(0, 10, 1001)
# gr.plot(Ax.axs[0, 0], t, np.sin(t), label="1")
# gr.plot(Ax.axs[1, 0], t, np.cos(t))
# gr.plot(Ax.axs[0, 0], t, np.log(t), label="2")
# gr.plot(Ax.axs[1, 0], t, np.exp(t))
# # gr.plot(Ax.axs[0, 2], t, np.sqrt(t))
# # gr.plot(Ax.axs[1, 2], t, np.square(t))
#
# Ax.set_props(sup_title="abc",
#              ax_title=[1, 2],
#              grid=[True, False],
#              limits=[
#                  None,
#                  [[-5, 5], [-1, 1]]
#              ],
#              show_fig=True)

# def outer(func):
#     def inner(x, y):
#         z = np.zeros(5)
#         for j in range(len(x)):
#             z[j] = func(x[j], y[j])
#         return z
#
#     return inner
#
#
# @outer
# def adder(x, y):
#     return x + y
#
# @outer
# def subtract(x, y):
#     return x - y
#
# @atexit.register
# def bye():
#     print("Bye")
#
#
# x = [0,1,2,3,4]
# y = [0,1,2,3,4]
# z = adder(x, y)


pass



import pandas as pd
from .. import uncertainties_math as un

load_csv = pd.read_csv


def load_csv_to_dict(file, *args, **kwargs) -> dict:
	table = pd.read_csv(file, *args, **kwargs)

	d = dict()
	for column in table.columns:
		d[column] = table[column].to_numpy()

	return d


def load_csv_to_dict_with_uncertainties(file, dev_str_identifier=" dev", *args, **kwargs) -> dict:
	d = load_csv_to_dict(file, *args, **kwargs)
	d_uncertainties = dict()

	for key in d.keys():
		if key.endswith(dev_str_identifier):
			key_wo_dev = key.split(dev_str_identifier)[0]
			d_uncertainties[key_wo_dev] = un.unumpy.uarray(d[key_wo_dev], d[key])

	return d_uncertainties



import numpy as np
from scipy.signal import firwin2, butter
from scipy.signal import windows

from liron_utils import graphics as gr
from liron_utils.signal_processing import analyze_window, analyze_filter

b = firwin2(numtaps=21, freq=[0, 0.1, 0.2, 1], gain=[1, 1, 0, 0])
# b = windows.get_window("hamming", 6)
# b /= b.sum()
a = 1

# b, a = butter(N=8, Wn=[0.3, 0.7], btype="bandpass")

# out = analyze_window(b, fs=None, plot=True)
out = analyze_filter(b, a, plot=True)

# Ax = gr.Axes(shape=(3, 1))
# Ax[0, 0].plot_impulse_response(b, a, n=50)
# Ax[1, 0].plot_frequency_response(b, a, one_sided=False, which="amp")
# Ax[2, 0].plot_frequency_response(b, a, one_sided=False, which="phase")
# Ax.set_props()



import numpy as np

"""
Prints:
	f"{a:.2fP}" -> "a±0.01"  (P = pretty-print ±)

Math functions 
	umath and unumpy has all the math functions that math has, but for arrays of ufloats

"""


def __repr__(self):
	# Not putting spaces around "+/-" helps with arrays of
	# Variable, as each value with an uncertainty is a
	# block of signs (otherwise, the standard deviation can be
	# mistaken for another element of the array).

	std_dev = self.std_dev  # Optimization, since std_dev is calculated

	# A zero standard deviation is printed because otherwise,
	# ufloat_fromstr() does not correctly parse back the value
	# ("1.23" is interpreted as "1.23(1)"):

	if std_dev:
		std_dev_str = repr(std_dev)
	else:
		std_dev_str = '0'

	return "%r±%s" % (self.nominal_value, std_dev_str)


try:
	import uncertainties

	uncertainties.core.AffineScalarFunc.__repr__ = __repr__
except ImportError:
	pass


def is_unumpy(arr):
	import uncertainties

	try:
		return any([isinstance(arr[i], uncertainties.core.AffineScalarFunc) for i in range(len(arr))])
	except TypeError:
		return False


def is_ufloat(x):
	return hasattr(x, "std_dev")


def to_numpy(x, xerr=None):
	"""
	Convert unumpy->numpy and ufloat->float. If already numpy, return as is.
	Args:
		x ():           unumpy or ufloat. In this case, don't need to provide xerr as it is already contained in x
		xerr ():        Needed only in case x is numpy instead of unumpy

	Returns:

	"""
	import uncertainties
	from uncertainties import unumpy

	def unumpy_to_numpy(arr):
		val = unumpy.nominal_values(arr)
		dev = unumpy.std_devs(arr)
		# if not np.any(dev):
		# 	dev = None
		return val, dev

	def ufloat_to_float(x):
		val = uncertainties.nominal_value(x)
		dev = uncertainties.std_dev(x)
		# if dev == 0:
		# 	dev = None
		return val, dev

	if is_unumpy(x):
		x, xerr = unumpy_to_numpy(x)

	elif is_ufloat(x):
		x, xerr = ufloat_to_float(x)

	return x, xerr


def from_numpy(x, xerr=0):
	from uncertainties import ufloat, unumpy

	if type(x) is float:
		return ufloat(x, xerr)
	else:
		return unumpy.uarray(x, xerr)


def val(x, xerr=None):
	return to_numpy(x, xerr)[0]


def dev(x, xerr=None):
	return to_numpy(x, xerr)[1]


def uncertainty(x, xerr=None):
	return dev(x, xerr) / val(x, xerr)


def make_independent(x, xerr=None):
	from uncertainties import unumpy

	x, xerr = to_numpy(x, xerr)

	return unumpy.uarray(x, xerr)


def histogram(x, xerr=None, **kwargs):
	x, xerr = to_numpy(x, xerr)

	density = kwargs.pop("density", False)

	hist, bins = np.histogram(x, **kwargs)

	hist_err = np.zeros(len(hist))
	for i in range(len(hist)):
		in_bin = (bins[i] <= x) & (x < bins[i + 1])
		hist_err[i] = np.sqrt(np.sum(xerr[in_bin] ** 2))

	hist = from_numpy(hist, hist_err)

	if density:
		hist /= np.sum(hist)

	return hist, bins



def get_pixel_color(modal=True):
	import pyautogui
	from pynput import mouse, keyboard
	from threading import Event
	
	done = Event()  # Event to block until user clicks or presses Esc

	def on_click(x, y, button, pressed):
		if pressed:
			rgb = pyautogui.screenshot().getpixel((x, y))
			rgb_norm = tuple([v / 255 for v in rgb])
			hex = "#{:02x}{:02x}{:02x}".format(*rgb).upper()
			print(f"(x, y)=({x}, {y}) | "
			      f"RGB=({rgb_norm[0]:.3f}, {rgb_norm[1]:.3f}, {rgb_norm[2]:.3f}) | "
			      f"HEX={hex}")
			done.set()
			listener_mouse.stop()
			listener_keyboard.stop()

	def on_press(key):
		if key == keyboard.Key.esc:
			print("Operation cancelled by user.")
			done.set()
			listener_mouse.stop()
			listener_keyboard.stop()

	print("Click to pick a pixel color or press Esc to cancel...")

	listener_mouse = mouse.Listener(on_click=on_click)
	listener_keyboard = keyboard.Listener(on_press=on_press)
	listener_mouse.start()
	listener_keyboard.start()

	if modal:
		done.wait()  # Block until the event is set



from liron_utils.pure_python import Logger

with Logger(
		min_level_file=Logger.NAME2LEVEL.INFO,
		min_level_console=Logger.NAME2LEVEL.DEBUG) as logger:
	logger.debug("This is a debug message. It will be printed to the console, but not to the log file.")
	logger.info("This is an info message.")
	logger.warning("This is a warning message.")
	logger.error("This is an error message.")
	logger.critical("This is a critical message.")

	x = 1 / 0  # This will raise a ZeroDivisionError and be automatically logged as an error



import os
import platform
import subprocess
import shutil

move_file = os.rename
remove_file = os.remove


def natural_sort(seq, key=None, reverse=False, alg=None):
	"""
	Sorts an iterable naturally.

	Parameters
	----------
	seq : iterable
	    The input to sort.

	key : callable, optional
	    A key used to determine how to sort each element of the iterable.
	    It is **not** applied recursively.
	    It should accept a single argument and return a single value.

	reverse : {{True, False}}, optional
	    Return the list in reversed sorted order. The default is
	    `False`.

	alg : ns enum, optional
	    This option is used to control which algorithm `natsort`
	    uses when sorting. For details into these options, please see
	    the :class:`ns` class documentation. The default is `ns.INT`.

	Returns
	-------
	out: list
	    The sorted input.

	See Also
	--------
	natsort_keygen : Generates the key that makes natural sorting possible.
	realsorted : A wrapper for ``natsorted(seq, alg=ns.REAL)``.
	humansorted : A wrapper for ``natsorted(seq, alg=ns.LOCALE)``.
	index_natsorted : Returns the sorted indexes from `natsorted`.
	os_sorted : Sort according to your operating system's rules.

	Examples
	--------
	Use `natsorted` just like the builtin `sorted`::

	    >>> a = ['num3', 'num5', 'num2']
	    >>> natsorted(a)
	    ['num2', 'num3', 'num5']

	"""

	import natsort

	if alg is None:
		alg = natsort.ns.DEFAULT

	return natsort.natsorted(seq=seq, key=key, reverse=reverse, alg=alg)


def mkdirs(dirs, *args, **kwargs):
	"""
	Create (possibly multiple) directories

	Parameters
	----------
	dirs :
	args :
	kwargs :

	Returns
	-------

	"""

	kwargs = {"exist_ok": True} | kwargs

	os.makedirs(dirs, *args, **kwargs)


def rmdir(dir):
	try:
		os.rmdir(dir)
	except FileNotFoundError:
		pass


def copy(src, dst, overwrite=True, symlink=None):
	"""
	Copy file(s) or directories, or create symbolic links.

	Parameters
	----------
	src : str | list[str]
		Source file(s) or directory(ies).
	dst : str | list[str]
		- If 'src' is a string, 'dst' can be a target path or directory.
		- If 'src' is a list, 'dst' must be either a directory or a list of same length.
	overwrite : bool, default=True
		Overwrite destination if it exists.
	symlink : bool or None, default=None
		If True, create symbolic links instead of copying files/directories.
		If None, follow src's type (if `src` is a symlink, create a symlink; otherwise, copy).
	"""

	# convert to list
	if isinstance(src, str):
		src = [src]
	if isinstance(dst, str):
		dst = [dst]

	if len(src) == 1:  # src is a single file or directory
		if len(dst) > 1:
			raise ValueError("Cannot copy file/directory to multiple destinations.")
	else:  # len(src) > 1:
		if len(dst) == 1:  # dst is a directory
			dst = [os.path.join(dst[0], os.path.basename(s)) for s in src]
		else:  # dst is a list of destinations
			if len(src) != len(dst):
				raise ValueError("When copying multiple files, `dst` must be a list of the same length as `src`.")

	symlink = [symlink] * len(src)

	def create_symlink(src, dst, target_is_directory=False):
		"""Create a symbolic link to the source file or directory."""
		# Remove existing destination if exists
		if src == dst:
			raise ValueError("Source and destination cannot be the same.")
		if os.path.isdir(dst):
			raise ValueError(f"Directory '{dst}' is not empty.")
		elif os.path.exists(dst):
			os.remove(dst)

		os.symlink(src, dst, target_is_directory=target_is_directory)

	for s, d, lnk in zip(src, dst, symlink):
		if not overwrite and os.path.exists(d):  # Skip if not overwriting and destination exists
			continue

		if os.path.islink(s):  # If src is a symlink, resolve it
			s = os.readlink(s).replace("\\\\?\\", "")
			if lnk is True:
				pass
			elif lnk is None:
				lnk = True
			else:
				lnk = False
		if os.path.islink(d):  # If dst is a symlink, resolve it
			d = os.readlink(d).replace("\\\\?\\", "")

		if os.path.isdir(s):  # copy directory
			mkdirs(os.path.dirname(d))  # Ensure the parent directory exists
			if lnk:
				create_symlink(s, d, target_is_directory=True)
			else:
				shutil.copytree(s, d, dirs_exist_ok=True)  # Copy directory recursively

		elif os.path.isfile(s):  # copy file
			if not os.path.splitext(d)[1]:  # If destination is a directory (i.e., has no extension), append filename
				d = os.path.join(d, os.path.basename(s))
			mkdirs(os.path.dirname(d))  # Ensure the parent directory exists
			if lnk:
				create_symlink(s, d, target_is_directory=False)
			else:
				shutil.copy2(s, d)  # Copy file with metadata

		else:
			raise ValueError(f"Source {s} is neither a file nor a directory.")


def open_file(file):
	if platform.system() == "Windows":
		os.startfile(file)
	elif platform.system() == "Darwin":
		subprocess.Popen(["open", file])
	else:
		subprocess.Popen(["xdg-open", file])



from typing import OrderedDict, Iterable, Iterator, TypeVar, Generic
import threading
import pandas as pd

T = TypeVar("T")


class MetaDict(type, Generic[T]):
    def __iter__(cls) -> Iterator[str]:
        for name, value in cls.__dict__.items():
            if not name.startswith("_"):
                yield name

    def __getitem__(cls, key) -> T:
        return getattr(cls, key)

    def keys(cls) -> list[str]:
        return list(iter(cls))

    def values(cls) -> list[T]:
        return [cls[k] for k in cls]

    def items(cls) -> list[tuple[str, T]]:
        return [(k, cls[k]) for k in cls]


class dict_(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def DL_to_LD(DL: dict) -> list:
    """
    Convert dict of lists to list of dicts

    Args:
        DL ():      Dict of lists

    Returns:

    """

    try:
        LD = pd.DataFrame(DL).to_dict(orient="records")
    except ValueError:  # If dict values are scalars
        LD = None

    return LD


def LD_to_DL(LD: list) -> dict:
    """
    Convert list of dicts to dict of lists

    Args:
        LD ():      List of dicts

    Returns:

    """

    DL = pd.DataFrame(LD).to_dict(orient="list")
    return DL


class NamedQueue:
    def __init__(self, max_size=0):
        """Thread-safe queue with a maximum size.
        If the queue is full, the oldest item is removed when a new item is added."""
        self.maxsize = max_size
        self.queue = OrderedDict()
        self._lock = threading.Lock()

    def enqueue(self, name, item):
        """Add an item to the queue. If the queue is full, remove the oldest item."""
        with self._lock:
            if name in self.queue:  # item already exists
                del self.queue[name]
            elif 0 < self.maxsize <= len(self.queue):  # queue is full
                self.queue.popitem(last=False)  # FIFO order
            self.queue[name] = item  # add new item

    def dequeue(self):
        """Remove and return the oldest item from the queue. If the queue is empty, return None."""
        with self._lock:
            if self.queue:  # queue is not empty
                return self.queue.popitem(last=False)
            return None

    def remove(self, name):
        """Remove an item from the queue by its name."""
        with self._lock:
            return self.queue.pop(name)

    def update(self, name, item):
        """Update an item in the queue by its name. If the item does not exist, it is added."""
        with self._lock:
            if name in self.queue:
                self.queue[name] = item
            else:
                self.enqueue(name, item)

    def __contains__(self, name):
        """Check if an item with the given name is in the queue."""
        with self._lock:
            return name in self.queue

    def __len__(self):
        return len(self.queue)

    def __getitem__(self, name):
        """Get an item from the queue by its name."""
        with self._lock:
            return self.queue[name]

    def __repr__(self):
        """Return a string representation of the queue."""
        with self._lock:
            return f"Queue({list(self.queue.items())})"

    def __iter__(self):
        """Return an iterator over the queue."""
        with self._lock:
            return iter(self.queue.items())

    def __next__(self):
        """Return the next item in the queue."""
        with self._lock:
            if self.queue:
                return self.queue.popitem(last=False)
            else:
                raise StopIteration

    def keys(self):
        """Return the keys of the queue."""
        with self._lock:
            return self.queue.keys()

    def values(self):
        """Return the values of the queue."""
        with self._lock:
            return self.queue.values()

    def items(self):
        """Return the items of the queue."""
        with self._lock:
            return self.queue.items()

    def clear(self):
        """Clear the queue."""
        with self._lock:
            self.queue.clear()



from __future__ import annotations

from manim import *
import numpy as np
import scipy.special


class PlotFunction(Scene):
    def construct(self):
        c_grid = ComplexPlane()
        moving_c_grid = ComplexPlane(
            faded_line_style={
                "stroke_color":   GREY,
                "stroke_width":   1,
                "stroke_opacity": .25,
            },
            faded_line_ratio=5
        )

        # c_grid.set_stroke(BLUE_E, 1)
        c_grid.add_coordinates()

        moving_c_grid.prepare_for_nonlinear_transform()

        self.play(
            Create(c_grid, run_time=2, lag_ratio=0.1)
        )
        self.wait(0.5)

        func_text = Tex(r"$f(z)$").to_corner(UP + LEFT)

        self.play(
            FadeIn(moving_c_grid),
            FadeIn(func_text, shift=UP),
        )
        self.wait(1.5)

        self.play(
            moving_c_grid.animate.apply_complex_function(func),
            run_time=6,
        )
        self.wait(4)


if __name__ == "__main__":
    func = lambda z: 2*z

    h = PlotFunction()
    h.construct()
    h.renderer.file_writer.finish()
    pass



from .csv import *
from .docx import *
from .files import *
from .json import *
from .pdf import *

import os
import sys

__all__ = [s for s in dir() if not s.startswith('_')]

try:
    MAIN_FILE_DIR = os.path.split(sys.modules['__main__'].__file__)[0]
except AttributeError:
    MAIN_FILE_DIR = ''



import sys
import inspect

is_debugger = sys.gettrace() is not None


def class_vars(cls):
    """Return the class variables of a class."""
    all_vars = dict()
    for name in dir(cls):
        # Filter out magic methods
        if name.startswith('__'):
            continue
        try:
            attr = getattr(cls, name)
            # Check if it's a property or variable
            if not inspect.ismethod(attr) and not inspect.isfunction(attr):
                all_vars[name] = attr
        except Exception:
            pass
    return all_vars


def is_notebook():
    try:
        from IPython import get_ipython
        if get_ipython() is not None:
            return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
        return False
    except ImportError:
        return False



import warnings
import sys
import functools
import multiprocessing as mp
import threading
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed

from .progress_bar import tqdm_

NUM_CPUS = mp.cpu_count()
NUM_PROCESSES_TO_USE = NUM_CPUS
NUM_THREADS_TO_USE = 20


def parallel_map(func: callable,
		iterable: Iterable,
		callback=None,
		error_callback=None,
		num_processes: int = NUM_PROCESSES_TO_USE,
		tqdm_kw=None,
		**kwargs) -> list:
	"""
	Run function 'func' in parallel.
	See qutip.parallel.parallel_map for reference.

	Notes
	-----
	- 'func' must be a global function (can't be nested inside another function).
	- parallel_map uses 'spawn' [1,2] by default in Windows, which starts a Python child process from scratch.
	  This means that everything not under an 'if __name__==__main__' block will be executed multiple times.
	- In UNIX we use 'fork'.

	References
	----------
	[1] https://stackoverflow.com/questions/64095876/multiprocessing-fork-vs-spawn/66113051#66113051
	[2] https://stackoverflow.com/questions/72935231/statements-before-multiprocessing-main-executed-multiple-times-python
	[3] https://superfastpython.com/multiprocessing-pool-issue-tasks

	Examples
	--------
	>>> import time
	>>> def func(iter, x, y):
	>>> 	time.sleep(1)
	>>> 	return (x + y) ** iter
	>>>
	>>> if __name__ == '__main__':
	>>> 	x = 1
	>>> 	y = 2
	>>> 	t0 = time.time()
	>>> 	out = parallel_map(func=func, iterable=range(100), num_processes=8, x=x, y=y)
	>>> 	print(out[:5])
	>>> 	print(f"time: {time.time() - t0}sec")

	Notes
	-----
	- If you get UserWarning: Can't pickle local object 'func.<locals>.func_partial', try to define 'func' in
	  the global scope.

	Parameters
	----------
	func :              callable
						The function to evaluate in parallel. The first argument is the changing value of each iteration
	iterable :          array_like
						First input argument for 'func'
	callback :          callable
						function to call after each iteration is done
	error_callback :    callable
						function to call if an error occurs
	num_processes :     int
						number of processes to use
	progress_bar :     bool
						if True, show a progress bar using tqdm
	desc :              str
						description for tqdm
	postfix :           callable
						postfix for tqdm description. If callable, should return a dict.
	tqdm_kw :           dict
						kwargs for tqdm.
	kwargs :            passed to func

	Returns
	-------
	list of 'func' outputs, organized by the order of 'iter'.

	"""
	if sys.platform == 'darwin':  # in UNIX 'fork' can be used (faster but more dangerous)
		Pool = mp.get_context('fork').Pool
	else:  # In Windows only 'spwan' is available
		Pool = mp.Pool
		mp.Process()

	if num_processes > NUM_CPUS:
		warnings.warn(f"Requested number of processes {num_processes} is larger than number of CPUs {NUM_CPUS}.\n"
		              f"For better performance, consider reducing 'num_processes'.", category=UserWarning)
	num_processes = min(num_processes, NUM_CPUS, len(iterable))

	with Pool(processes=num_processes) as pool:
		func_partial = functools.partial(func, **kwargs)  # pass kwargs to func

		out_async = [pool.apply_async(
				func=func_partial,
				args=(i,),
				callback=callback,
				error_callback=error_callback
		) for i in iterable]

		out = []

		for out_async_i in tqdm_(out_async, **tqdm_kw):
			try:
				out += [out_async_i.get()]

			except KeyboardInterrupt as e:
				raise e

			except Exception as e:
				warnings.warn(f"Exception at index {out_async_i}: {e}")

	return out


def parallel_threading(func: callable,
		iterable: Iterable,
		lock: bool = False,
		num_threads: int = NUM_THREADS_TO_USE,
		tqdm_kw=None,
		**kwargs) -> list:
	"""
	Run function 'func' in parallel using threads.

	Parameters
	----------
	func :          callable
					The function to evaluate using threads. The first argument is the changing value of each iteration.
	iterable :      array_like
					First input argument for 'func'
	lock :          bool
					Use a lock to prevent concurrent access to shared resources.
	num_threads :   int
					Number of threads to use.
	progress_bar : bool
					If True, show a progress bar using tqdm.
	desc :          str or callable
					Description for tqdm. If callable, should return a string.
	postfix :       callable
					Postfix for tqdm description. If callable, should return a dict.
	tqdm_kw :       dict
					kwargs for tqdm.
	kwargs :        Passed to func.

	Returns
	-------
	List of 'func' outputs, organized by the order of 'iterable'.
	"""
	if lock:
		lock = threading.Lock()

	def wrapped_func(index, item):
		try:
			if lock:
				with lock:
					result = func(item, **kwargs)
			else:
				result = func(item, **kwargs)
			return index, result
		except Exception as e:
			warnings.warn(f"Exception at index {index}: {e}")
			return index, None

	out = [None] * len(iterable)
	with ThreadPoolExecutor(max_workers=num_threads) as executor:
		futures = [executor.submit(wrapped_func, i, item) for i, item in enumerate(iterable)]

		if tqdm_kw is None:
			tqdm_kw = dict()
		tqdm_kw = dict(total=len(futures)) | tqdm_kw

		for future in tqdm_(as_completed(futures), **tqdm_kw):
			index, result = future.result()
			out[index] = result

	return out



import os
import numpy as np
import scipy.signal
from .base import array


def resample(y: np.ndarray, fs_old: int, fs_new: int, axis: int = -1, *args, **kwargs):
    if fs_old == fs_new:
        return y
    n = round(y.shape[axis] * fs_new / fs_old)
    return scipy.signal.resample(y, n, axis=axis, *args, **kwargs)


def audio_read(files: (str, list), fs_new: int = None, always_3d: bool = False, **kwargs):
    """
    Audio Reader

    Parameters
    ----------
    files       :   File names, specified either as a list or a string (in case of 1 signal).
    fs_new      :   Desired sample rate (if different from the original)
    always_3d   :   Always return signals array as a 3D array with dimensions (m, n, ch), where:
                        m - number of signals
                        n - number of frames
                        ch - number of channels (can be either 1 or 2)
    kwargs      :   Passed to <soundfile.read>

    Returns
    -------

    """
    import soundfile as sf

    files = np.atleast_1d(array(files))
    assert files.ndim == 1
    (channels, samplerate, frames) = [], [], []
    for i, f in enumerate(files):
        info = sf.info(f)
        channels.append(info.channels), samplerate.append(info.samplerate), frames.append(info.frames)
    (channels, samplerate, frames) = np.unique(channels), np.unique(samplerate), np.unique(frames)
    assert channels.size == 1, 'Inconsistent number of channels.'
    assert samplerate.size == 1, 'Inconsistent sample rates.'
    assert frames.size == 1, 'Inconsistent signal lengths.'

    if 'start' not in kwargs.keys():
        kwargs['start'] = 0
    if 'stop' not in kwargs.keys():
        kwargs['stop'] = frames[0]
    frames = kwargs['stop'] - kwargs['start']

    info = sf.info(files[0])
    M = np.zeros((len(files), frames, info.channels))

    # Read
    for i, f in enumerate(files):
        (M[i, :, :], _) = sf.read(f, always_2d=True, **kwargs)

    # Resample
    if fs_new is not None:
        M = resample(M, info.samplerate, fs_new, axis=1)

    if not always_3d:
        M = M.squeeze()
    return M, samplerate[0]


def audio_write(files, M, fs, fs_save=None, mode='w', **kwargs):
    """
    Audio Writer

    Parameters
    ----------
    files       :   str or list
        File names, specified either as a list or a string (in case of 1 file).
    M           :   array_like
        Signals 3D array, with dimensions (m, n, ch), where:
            m - number of signals
            n - number of frames
            ch - number of channels (can be either 1 or 2)
        If M is given as 1D, M is interpreted as a single channel.
        If M is given as 2D, then:
            If it has dimensions (n, 2), it is interpreted as single stereo signal.
            If it has dimensions (m, n), it is interpreted as m different signals.
    fs          :   int
        Original sample rate [Hz]
    fs_save     :   int, optional
        Desired sample rate to save signals (if different from the original) [Hz]
    mode        :   str {'w', 'a'}
        Save mode (write / append)
    kwargs      :   dict
        Passed to <soundfile.write>

    Returns
    -------

    """
    import soundfile as sf

    files = np.atleast_1d(array(files))
    M = array(M)
    if M.ndim == 1:
        M = M[np.newaxis, :, np.newaxis]
    elif M.ndim == 2:
        if M.shape[1] == 2:
            M = M[np.newaxis, :, :]
        else:
            M = M[:, :, np.newaxis]
    assert files.ndim == 1 and M.ndim == 3
    assert files.size == M.shape[0]

    # Resample
    if fs_save is None:
        fs_save = fs
    else:
        M = resample(M, fs, fs_save, axis=1)

    # Write
    for i, f in enumerate(files):
        if mode.lower() == 'a':
            if not os.path.isfile(f):
                with sf.SoundFile(f, 'w', samplerate=fs_save, **kwargs):
                    pass

            with sf.SoundFile(f, 'r+') as h:
                h._update_frames(h.seek(0, sf.SEEK_END))
                h.buffer_write(M, dtype='float64')

        elif mode.lower() == 'w+':
            sf.write(f, M[i, :, :].squeeze(), fs_save, **kwargs)

        else:
            raise ValueError('Invalid declaration of variable `mode`.')


def time_array(n: int, fs: int) -> np.ndarray:
    """
    Create time array range from 0 to (n-1)/fs with n samples.

    Parameters
    ----------
    n :     Length
    fs :    Sample rate

    Returns
    -------

    """
    t = np.linspace(0, (n - 1) / fs, n)
    return t


def time2samples(fs: int, *args):
    args = list(args)
    for i in range(len(args)):
        if args[i] is not None:
            args[i] = round(fs * args[i])
    if len(args) == 1:
        return args[0]
    return args



from typing import Iterable
import numpy as np
import scipy.signal
import matplotlib.pyplot as plt
import matplotlib.cm
import matplotlib.collections
import matplotlib.colors
import matplotlib.animation
from matplotlib.figure import Figure
from matplotlib.axes import Axes as Axes_plt

from .axes import _Axes
from ..uncertainties_math import to_numpy
from ..signal_processing.base import interp1


# TODO:
#   - add option to call gr.Axes.plot(ax, x, y) with ax=plt.Axes (add decorator to outer plotting functions),
#     and change class name to be Plot
#   - add standalone functions out of class gr.plot
#   - PyWavelet.dwt (wavelet transform) and plot_wavelet


class Axes(_Axes):
    def __init__(self,
            shape: tuple[int] = (1, 1),
            grid_layout: list[list[tuple]] = None,
            sharex: bool | str = False, sharey: bool | str = False,
            projection: str = None,
            layout: str = None,
            fig: Figure = None, axs: Axes_plt | Iterable[Axes_plt] = None,
            subplot_kw: dict = None, gridspec_kw: dict = None, **fig_kw):
        super().__init__(shape=shape, grid_layout=grid_layout, sharex=sharex, sharey=sharey, projection=projection,
                layout=layout, fig=fig, axs=axs, subplot_kw=subplot_kw, gridspec_kw=gridspec_kw, **fig_kw)

    def plot(self,
            x, y=None, z=None,
            **plot_kw):

        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, x=x, y=y, z=z)
        def _plot(ax: Axes_plt,
                x, y=None, z=None,
                **plot_kw):
            """
            2D plot of y=f(x)

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            x : array_like
                X-axis data values.
            y : array_like, optional
                Y-axis data values. If None, x is treated as y and x becomes range(len(y)).
            z : array_like, optional
                Z-axis data values for 3D plots.
            **plot_kw : dict
                Additional keyword arguments passed to matplotlib.pyplot.plot.

            Returns
            -------
            list of Line2D
                A list of Line2D objects representing the plotted data.
            """

            args = [x]
            if y is not None:
                args += [y]
            if z is not None:
                args += [z]

            return ax.plot(*args, **plot_kw)

        return _plot()

    def plot_vlines(self,
            x=0, ymin=0, ymax=1,
            **plot_kw):

        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, x=x, ymin=ymin, ymax=ymax)
        def _plot_vlines(ax: Axes_plt,
                x=0, ymin=0, ymax=1,
                **plot_kw):
            """
            Plot vertical lines.

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            x : array_like or scalar, default 0
                x positions of the vertical lines.
            ymin : float, default 0
                Minimum y position of the lines (in data coordinates).
            ymax : float, default 1
                Maximum y position of the lines (in data coordinates).
            **plot_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.axvline.

            Returns
            -------
            list of Line2D
                A list of Line2D objects representing the plotted vertical lines.
            """
            x = np.atleast_1d(x)

            label = None
            if "label" in plot_kw:
                label = plot_kw.pop("label")

            lines = []
            for i, xx in enumerate(x):
                line = ax.axvline(x=xx, ymin=ymin, ymax=ymax,
                        label=label if i == x.shape[0] - 1 else "_nolabel_",
                        **plot_kw)
                lines.append(line)

            return lines

        return _plot_vlines()

    def plot_hlines(self,
            y=0, xmin=0, xmax=1,
            **plot_kw):

        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, y=y, xmin=xmin, xmax=xmax)
        def _plot_hlines(ax: Axes_plt,
                y=0, xmin=0, xmax=1,
                **plot_kw):
            """
            Plot horizontal lines.

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            y : array_like or scalar, default 0
                y positions of the horizontal lines.
            xmin : float, default 0
                Minimum x position of the lines (in data coordinates).
            xmax : float, default 1
                Maximum x position of the lines (in data coordinates).
            **plot_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.axhline.

            Returns
            -------
            list of Line2D
                A list of Line2D objects representing the plotted horizontal lines.
            """
            y = np.atleast_1d(y)

            label = None
            if "label" in plot_kw:
                label = plot_kw.pop("label")

            lines = []
            for i, yy in enumerate(y):
                line = ax.axhline(y=yy, xmin=xmin, xmax=xmax,
                        label=label if i == y.shape[0] - 1 else "_nolabel_",
                        **plot_kw)
                lines.append(line)

            return lines

        return _plot_hlines()

    def plot_errorbar(self,
            x, y=None, xerr=None, yerr=None,
            **errorbar_kw):

        @self._merge_kwargs("errorbar_kw", **errorbar_kw)
        @self._vectorize(cls=self, x=x, y=y, xerr=xerr, yerr=yerr)
        def _plot_errorbar(ax: Axes_plt,
                x, y=None, xerr=None, yerr=None,
                **errorbar_kw):
            """
            2D plot of y=f(x) with errorbars

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            x : array_like
                x-axis data values.
            y : array_like, optional
                y-axis data values. If None, x is treated as y and x becomes range(len(x)).
            xerr, yerr : array_like, optional
                Error bars in the x, y direction.
            **errorbar_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.errorbar.

            Returns
            -------
            ErrorbarContainer
                Container with the errorbar plot elements.
            """

            if y is None:
                assert xerr is None and yerr is None, "If y is not given, xerr and yerr should not be given."
                y = x
                x = np.arange(len(y))

            x, xerr = to_numpy(x, xerr)
            y, yerr = to_numpy(y, yerr)

            return ax.errorbar(x, y, xerr=xerr, yerr=yerr, **errorbar_kw)

        return _plot_errorbar()

    def plot_filled_error(self, ax: Axes_plt,
            x,
            y=None, yerr=None, n_std=2,
            y_low=None, y_high=None,
            **fill_between_kw):

        @self._merge_kwargs("fill_between_kw", **fill_between_kw)
        @self._vectorize(cls=self, ax=ax,
                x=x,
                y=y, yerr=yerr, n_std=n_std,
                y_low=y_low, y_high=y_high)
        def _plot_filled_error(ax: Axes_plt,
                x,
                y=None, yerr=None, n_std=2,
                y_low=None, y_high=None,
                **fill_between_kw):
            """
            Plot error using filled area

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            x : array_like
                x-axis data values.
            y : array_like, optional
                y-axis data values.
            yerr : array_like, optional
                Error values in the y direction.
            n_std : int, default 2
                Number of standard deviations to fill.
            y_low : array_like, optional
                Lower bounds of the filled area. If not given, will be calculated as y-n_std*yerr.
            y_high : array_like, optional
                Upper bounds of the filled area. If not given, will be calculated as y+n_std*yerr.
            **fill_between_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.fill_between.

            Returns
            -------
            PolyCollection
                The filled area between the curves.
            """

            # Data
            x, _ = to_numpy(x)
            if y is None:
                assert y_low is not None and y_high is not None, "(y, yerr) or (y_low, y_high) should be given."
            else:
                assert y_low is None and y_high is None, "(y, yerr) or (y_low, y_high) should be given."
                y, yerr = to_numpy(y, yerr)
                assert yerr is not None, "yerr should be given."

                y_low = y - n_std * yerr
                y_high = y + n_std * yerr

            return ax.fill_between(x, y_low, y_high,
                    **fill_between_kw)

        return _plot_filled_error()

    def plot_data_and_curve_fit(self,
            x, y, fit_fcn, xerr=None, yerr=None,
            p_opt=None, p_cov=None, n_std=2, interp_factor=20,
            curve_fit_plot_kw=None, **errorbar_kw):

        @self._merge_kwargs("errorbar_kw", **errorbar_kw)
        @self._vectorize(cls=self,
                x=x, y=y, fit_fcn=fit_fcn, xerr=xerr, yerr=yerr,
                p_opt=p_opt, p_cov=p_cov, n_std=n_std, interp_factor=interp_factor,
                curve_fit_plot_kw=curve_fit_plot_kw)
        def _plot_data_and_curve_fit(ax: Axes_plt,
                x, y, fit_fcn, xerr=None, yerr=None,
                p_opt=None, p_cov=None, n_std=2, interp_factor=20,
                curve_fit_plot_kw=None, **errorbar_kw):
            """
            2D scatter plot y=f(x) and curve fit.

            Parameters
            ----------
            ax :
            x, y : array_like
                Data points to be plotted.
            fit_fcn : callable
                Function to be fitted. The first argument of fit_fcn should be the independent variable (x).
            xerr, yerr : array_like, optional
                Deviations in 'x','y'. 'x','y' may also be sent as 'uncertainties' arrays, then 'xerr','yerr' are disregarded
            p_opt : Output parameter of scipy.optimize.curve_fit
            p_cov : Output parameter of scipy.optimize.curve_fit
            n_std : int, optional, default 2
                Number of standard deviations of confidence to be plotted with the fitted curve
            interp_factor : int, optional, default 20
                Interpolation factor for the fitted curve.
            curve_fit_plot_kw : dict, optional
            **errorbar_kw : dict, optional

            Returns
            -------

            Examples
            --------
                >>> import numpy as np
                >>> from liron_utils import graphics as gr
                >>> import scipy.optimize

                >>> N = 101
                >>> x = np.linspace(0, 10, N)
                >>> yerr = 5 * np.random.randn(N)
                >>> y = 2 * x ** 2 + 4 * x + 5 + yerr

                >>> def fit_fcn(x, a, b, c):
                >>>     return a * x ** 2 + b * x + c

                >>> (p_opt, p_cov) = scipy.optimize.curve_fit(fit_fcn, x, y)

                >>> Ax = gr.Axes()
                >>> Ax.plot_data_and_curve_fit(x, y, fit_fcn, yerr=yerr, p_opt=p_opt, p_cov=p_cov)
                >>> Ax.show_fig()
            """

            errorbar_kw = {
                              "label":  'Data',
                              "zorder": -1
                          } | errorbar_kw

            if curve_fit_plot_kw is None:
                curve_fit_plot_kw = dict()
            curve_fit_plot_kw = {
                                    "label": 'Curve fit'
                                } | curve_fit_plot_kw

            # Data
            x, xerr = to_numpy(x, xerr)
            y, yerr = to_numpy(y, yerr)
            p_opt, _ = to_numpy(p_opt)
            idx = np.argsort(x)
            x, y, xerr, yerr = x[idx], y[idx], xerr[idx], yerr[idx]

            self.plot_errorbar(x, y, xerr=xerr, yerr=yerr, **errorbar_kw)

            # Curve fit
            x_interp = interp1(x, interp_factor * len(x))
            fit_mid = fit_fcn(x_interp, *p_opt)

            ax.plot(x_interp, fit_mid, **curve_fit_plot_kw)

            # Confidence fill
            if p_opt is not None and p_cov is not None:
                p_err = np.sqrt(np.diag(p_cov))
                fit_low = np.ones(interp_factor * len(x)) * np.inf
                fit_high = np.ones(interp_factor * len(x)) * (-np.inf)

                for i in range(len(p_opt)):
                    p_opt_i = p_opt.copy()
                    p_opt_i[i] = p_opt[i] - n_std * p_err[i]
                    low = fit_fcn(x_interp, *p_opt_i)
                    p_opt_i[i] = p_opt[i] + n_std * p_err[i]
                    high = fit_fcn(x_interp, *p_opt_i)

                    fit_low = np.minimum(fit_low, np.minimum(low, high))
                    fit_high = np.maximum(fit_high, np.maximum(low, high))

                self.plot_filled_error(ax=ax, x=x_interp, y_low=fit_low, y_high=fit_high)

        return _plot_data_and_curve_fit()

    def plot_data_and_lin_reg(self,
            x, y, reg=None, xerr=None, yerr=None,
            reg_plot_kw=None, **errorbar_kw):

        @self._merge_kwargs("errorbar_kw", **errorbar_kw)
        @self._vectorize(cls=self,
                x=x, y=y, reg=reg, xerr=xerr, yerr=yerr,
                reg_plot_kw=reg_plot_kw)
        def _plot_data_and_lin_reg(ax: Axes_plt,
                x, y, reg=None, xerr=None, yerr=None,
                reg_plot_kw=None, **errorbar_kw):
            """
            2D scatter plot y=f(x) + linear regression.

            Examples:
                >>> import numpy as np
                >>> from scipy.stats import linregress
                >>> from liron_utils import graphics as gr

                >>> N = 100
                >>> x = np.arange(N)
                >>> y = 2*x + np.random.randn(N)
                >>> reg = linregress(x, y)

                >>> ax = gr.Axes()
                >>> ax.plot_data_and_lin_reg(x, y, reg)
                >>> ax.show_fig()

            Parameters
            ----------
            ax :
            x :
            y :                      f(x)
            reg :                    Output of scipy.stats.linregress
            xerr :                   Error in x
            yerr :                   Error in y
            reg_plot_kw :
            **errorbar_kw :

            Returns
            -------

            """

            errorbar_kw = {
                              "label": "Data"
                          } | errorbar_kw

            if reg_plot_kw is None:
                reg_plot_kw = dict()
            reg_plot_kw = {
                              "label": fr"{errorbar_kw['label']} linreg: slope={reg.slope:.3f}$\pm${reg.stderr:.3f}, $R^2$={reg.rvalue ** 2:.3f}"
                          } | reg_plot_kw

            x, xerr = to_numpy(x, xerr)
            y, yerr = to_numpy(y, yerr)

            self.plot_errorbar(x, y, xerr=xerr, yerr=yerr,
                    **errorbar_kw)  # TODO: need to change to _plot_errorbar

            if reg is not None:
                tmp = [reg.slope * i + reg.intercept for i in x]
                self.plot(x, tmp, **reg_plot_kw)

        return _plot_data_and_lin_reg()

    def plot_line_collection(self,
            x: np.ndarray, y: np.ndarray, arr: np.ndarray,
            colorbar_kw=None, **LineCollection_kw):

        @self._vectorize(cls=self, x=x, y=y, arr=arr)
        def _plot_line_collection(ax: Axes_plt,
                x: np.ndarray, y: np.ndarray, arr: np.ndarray,
                colorbar_kw=None, **LineCollection_kw):
            """
            Create a line collection with colors mapped to array values.

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            x, y : np.ndarray
                x- and y-coordinates of the line segments.
            arr : np.ndarray
                Array of values to map to colors for each segment.
            colorbar_kw : dict, optional
                Additional keyword arguments passed to colorbar.
            **LineCollection_kw : dict
                Additional keyword arguments passed to matplotlib.collections.LineCollection.

            Returns
            -------
            tuple
                A tuple containing (LineCollection, colorbar) objects.
            """
            if colorbar_kw is None:
                colorbar_kw = dict()

            points = np.array([x, y]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)

            norm = plt.Normalize(vmin=np.min(arr), vmax=np.max(arr))

            # cmap = matplotlib.colors.ListedColormap(['r', 'g', 'b'])
            lc = matplotlib.collections.LineCollection(segments, norm=norm, **LineCollection_kw)
            lc.set_array(arr)

            line = ax.add_collection(lc)

            cbar = ax.figure.colorbar(line, ax=ax, **colorbar_kw)

            return line, cbar

        return _plot_line_collection(colorbar_kw=colorbar_kw, **LineCollection_kw)

    def plot_fft(self,
            x, fs=1.0, n=None,
            *,
            one_sided=True,
            dB=False,
            eps=1e-20,
            which="power",
            normalize=False,
            input_time=True,
            **plot_kw):

        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, x=x, fs=fs, n=n, one_sided=one_sided, dB=dB, eps=eps, which=which,
                normalize=normalize, input_time=input_time)
        def _plot_fft(ax: Axes_plt,
                x, fs=1.0, n=None,
                *,
                one_sided=True,
                dB=False,
                eps=1e-20,
                which="power",
                normalize=False,
                input_time=True,
                **plot_kw):
            """
            Plot the magnitude spectrum of the FFT of a signal.

            Parameters
            ----------
            x : np.ndarray
                Input signal (1D array).
            fs : float, optional, default 1.0
                Sampling frequency (Hz).
            n : int, optional
                Number of points for FFT. If None, uses len(y).
            one_sided : bool, optional, default True
                If True, plots only the positive frequencies.
            dB : bool, optional, default False
                If True, plot in decibels.
            eps : float, optional
                Renormalization constant added to logarithmic plot
            which : str, optional, default "power"
                Choose what to plot: "amp", "power", or "phase".
            normalize : bool, optional, default False
                If True, normalize the transformed signal to have a maximal value of 1.
            input_time : bool, optional, default True
                If True, assumes the input signal is in the time domain (otherwise assumes frequency domain).
            **plot_kw : dict
                Additional keyword arguments for the plot.

            Returns
            -------
            Line2D
                The line object for the FFT plot.
            """

            x = np.asarray(x)
            if n is None:
                n = x.shape[0]

            if input_time:
                X = np.fft.fft(x, n=n)
            else:
                X = x.copy()

            if normalize:
                X /= X.max(axis=0)

            freqs = np.fft.fftfreq(n=n, d=1 / fs)

            if one_sided:
                X = X[:n // 2]
                freqs = freqs[:n // 2]
            else:
                X = np.fft.fftshift(X, axes=0)
                freqs = np.fft.fftshift(freqs)

            which = which.lower()
            if which == "amp":
                ydata = np.abs(X)
                ylabel = "Amplitude"
            elif which == "power":
                ydata = np.abs(X) ** 2
                ylabel = "Power"
            elif which == "phase":
                ydata = np.degrees(np.unwrap(np.angle(X)))
                ylabel = "Phase [deg]"
            # ticks = np.arange(-180, 181, 45)
            # ax.set_yticks(ticks)
            # tick_labels = [rf"${t}^\circ$" for t in ticks]
            # ax.set_yticklabels(tick_labels)

            else:
                raise ValueError(f"which must be one of 'amp', 'power', or 'phase'. Got: {which}")

            if dB and which in ("amp", "power"):
                ydata = 10 * np.log10(ydata + eps)
                ylabel += " [dB]"

            line = ax.plot(freqs, ydata, **plot_kw)

            if ax.get_xlabel() == "":
                ax.set_xlabel("Frequency [Hz]")
                if fs == 1.0:
                    ax.set_xlabel("Frequency [normalized]")
            if ax.get_ylabel() == "":
                ax.set_ylabel(ylabel)

            return (X, freqs), line

        return _plot_fft()

    def plot_impulse_response(self,
            b, a=1,
            dt=1, t=None, n=None,
            **plot_kw):
        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, b=b, a=a, dt=dt, t=t, n=n)
        def _plot_impulse_response(ax: Axes_plt,
                b, a=1,
                dt=1, t=None, n=None,
                **plot_kw):
            """
            Plot the impulse response of a linear time-invariant (LTI) system defined by its numerator (b) and
            denominator (a) coefficients.

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            b : array_like
                Numerator coefficients of the LTI system.
            a : array_like, optional, default 1
                Denominator coefficients of the LTI system.
            dt : float, optional, default 1
                Sampling time interval. If None, the system is treated as continuous-time.
            t : array_like, optional
                Time values at which to compute the impulse response. If None, uses n and dt to generate time values.
            n : int, optional
                Number of samples for discrete-time systems. Required if t is None and dt is not None.
            **plot_kw : dict
                Additional keyword arguments for the plot.

            Returns
            -------

            """

            b, a = np.atleast_1d(b), np.atleast_1d(a)
            if len(b) > len(a):
                a = np.pad(a, (0, len(b) - len(a)), 'constant', constant_values=0)
            elif len(a) > len(b):
                b = np.pad(b, (0, len(a) - len(b)), 'constant', constant_values=0)

            if dt is None:  # Continuous-time system
                system = scipy.signal.lti(b, a)
                if t is None:
                    raise ValueError("t should be given for continuous-time system.")
                t, h = scipy.signal.impulse(system, T=t)
                line = ax.plot(t, h, **plot_kw)

            else:  # Discrete-time system
                system = scipy.signal.dlti(b, a, dt=dt)
                if t is None:
                    if n is None:
                        raise ValueError("Either t or n should be given.")
                    t = np.arange(0, n * dt, dt)
                t, h = scipy.signal.dimpulse(system, n=len(t))
                h = np.squeeze(h)
                line = ax.stem(t, h, **plot_kw)

            return (h, t), line

        return _plot_impulse_response()

    def plot_frequency_response(self,
            b, a=1, fs=1.0, worN=512,
            *,
            one_sided=True,
            dB=False,
            eps=1e-20,
            which="amp",
            normalize=False,
            **plot_kw):
        @self._merge_kwargs("plot_kw", **plot_kw)
        @self._vectorize(cls=self, b=b, a=a, fs=fs, worN=worN,
                one_sided=one_sided, dB=dB, eps=eps, which=which, normalize=normalize)
        def _plot_frequency_response(ax: Axes_plt,
                b, a=1, fs=1.0, worN=512,
                *,
                one_sided=True,
                dB=False,
                eps=1e-20,
                which="amp",
                normalize=False,
                **plot_kw):
            """
            Plot the frequency response of a linear time-invariant (LTI) system defined by its numerator (b) and
            denominator (a) coefficients.

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            b :             array_like
                Numerator coefficients of the LTI system.
            a :             array_like, optional, default 1
                Denominator coefficients of the LTI system.
            fs :            float, optional, default 1.0
                Sampling frequency (Hz). If None, the system is treated as continuous-time.
            worN :          int, optional, default 512
                Number of frequency points to compute.
            one_sided :     bool, optional, default True
                If True, plots only the positive frequencies.
            dB :            bool, optional, default False
                If True, plot in decibels.
            eps :           float, optional, default 1e-20
                Renormalization constant added to logarithmic plot
            which :         str, optional, default "amp"
                Choose what to plot: "amp", "power", or "phase".
            normalize :     bool, optional, default False
                If True, normalize the transformed signal to have a maximal value of 1.
            **plot_kw : dict
                Additional keyword arguments for the plot.

            Returns
            -------

            """

            if fs is None:  # Continuous-time system
                w, h = scipy.signal.freqs(b=b, a=a,
                        worN=worN)

            else:  # Discrete-time system
                w, h = scipy.signal.freqz(b=b, a=a,
                        fs=2 * np.pi * fs,
                        worN=worN,
                        whole=not one_sided)

            freqs = w / (2 * np.pi) - 0.5

            _, line = Axes(axs=ax).plot_fft(
                    x=h, fs=fs,
                    n=2 * worN if one_sided else worN,
                    one_sided=one_sided, dB=dB, eps=eps, which=which, normalize=normalize,
                    input_time=False,
                    **plot_kw)[0, 0]

            return (h, freqs), line

        return _plot_frequency_response()

    def plot_specgram(self,
            y: np.ndarray, fs: int,
            **specgram_kw):

        @self._merge_kwargs("specgram_kw", **specgram_kw)
        @self._vectorize(cls=self, y=y, fs=fs)
        def _plot_specgram(ax: Axes_plt,
                y: np.ndarray, fs: int,
                **specgram_kw):
            """
            Plot spectrogram

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on.
            y : np.ndarray
                Data, given as 1D array with respect to time.
            fs : int
                Sample rate.
            **specgram_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.specgram.

            Returns
            -------
            tuple
                A tuple containing (spectrum, freqs, t, im) from ax.specgram.
            """

            specgram_out = ax.specgram(y, Fs=fs,
                    **specgram_kw)  # todo: add option for log frequency mapping using librosa.feature.melspectrogram()
            spectrum, freqs, t, im = specgram_out

            scaling = {0: "", 3: "K", 6: "M", 9: "G", 12: "T"}  # todo: use siprefix.si_format
            scale = 0
            if fs != 1:
                scale = 3 * (int(np.log10(freqs[-1])) // 3)

                extent = im.get_extent()
                extent = (extent[0], extent[1], extent[2] / 10 ** scale, extent[3] / 10 ** scale)
                im.set_extent(extent=extent)

            if ax.get_title() == "":
                ax.set_title("Spectrogram")
            if ax.get_xlabel() == "":
                ax.set_xlabel("Time [sec]")
            if ax.get_ylabel() == "":
                ax.set_ylabel(f"Frequency [{scaling[scale]}Hz]")

            ax.figure.colorbar(im, ax=ax)

            return specgram_out

        return _plot_specgram()

    def plot_surf(self,
            x, y, z,
            **plot_surface_kw):

        @self._merge_kwargs("plot_surface_kw", **plot_surface_kw)
        @self._vectorize(cls=self, x=x, y=y, z=z)
        def _plot_surf(ax: Axes_plt,
                x, y, z,
                **plot_surface_kw):
            """
            3D surface plot of z=f(x,y)

            Parameters
            ----------
            ax : matplotlib.axes.Axes
                The axes to plot on (must have 3D projection).
            x : array_like
                X-coordinates. Can be 1D or 2D. If 1D, will automatically apply meshgrid.
            y : array_like
                Y-coordinates. Can be 1D or 2D. If 1D, will automatically apply meshgrid.
            z : array_like or callable
                Z-coordinates as 2D array or function z=f(x,y).
            **plot_surface_kw : dict
                Additional keyword arguments passed to matplotlib.axes.Axes.plot_surface.

            Returns
            -------
            Poly3DCollection
                The 3D surface plot object.
            """

            assert hasattr(ax, "plot_surface"), "Axes does not have a plot_surface attribute. " \
                                                "make sure that you created an axes with projection='3d'"

            X, Y, Z = x, y, z
            if x.ndim == 1:
                X, Y = np.meshgrid(x, y)
            if callable(z):
                Z = z(X, Y)
            if np.all(Z.shape == np.flip(X.shape)):
                Z = Z.T

            ax.plot_surface(X, Y, Z, **plot_surface_kw)

        # ax.figure.colorbar(matplotlib.cm.ScalarMappable(), ax=ax)

        return _plot_surf()

    def plot_contour(self,
            x, y, z, contours,
            *args, **kwargs):

        @self._vectorize(cls=self, x=x, y=y, z=z, contours=contours)
        def _plot_contour(ax: Axes_plt,
                x, y, z, contours,
                *args, **kwargs):
            """
            Contour plot of scalar field z=f(x,y)

            Parameters
            ----------
            ax :
            x :
            y :
            z :
            contours :
            args, kwargs :
             :

            Returns
            -------

            """

            cs = ax.contour(x, y, z, contours, *args, **kwargs)
            ax.clabel(cs, inline=True, fontsize=10)
            return cs

        return _plot_contour(*args, **kwargs)

    def plot_animation(self, axs: (Axes_plt, np.ndarray[Axes_plt]),
            func: callable = None,
            n_frames: int = None,
            data: list[np.ndarray] = None,
            data_instance: list = None,
            titles: list = None,
            *args, **kwargs):
        """
        Plot animation

        Examples
        --------
            >>> import numpy as np
            >>> from liron_utils import graphics as gr

            >>> nimages = 10
            >>> images = np.random.random((nimages))
            >>> Ax = gr.Axes()
            >>> Ax.plot_animation(images)
            >>> Ax.save_fig("test.gif")

        Parameters
        ----------
        axs :               Axes or list[Axes]
                            All axes should be of the same figure
        func :              callable, optional
                            Function to be called for each frame. If not given, will use the default `update_data`.
                            `func` should return a list of artists to draw in the new frame (same as `data_instance`).
        n_frames :	    int, optional
                            Number of frames to be plotted. If not given, will use len(data[0])
        data :              array_like, optional
                            The data to be plotted, given as list of size len(axs) of:
                                - (image) a 4D array of size [#n_frames, x, y]
                                - (plot)  a 4D array of size [#n_frames, 2, xy]
        data_instance   :   array_like of matplotlib objects, optional
                            Image/line/etc. handle to use in case user wants some pre-defined properties.
                            First axis should be of size 'len(axs)'
        titles :            list or function handle, optional
                            Axis titles. Can be given as:
                                - List of changing titles
                                - function handle whose input argument is iterable and outputs the title
        args, kwargs :      sent to matplotlib.animation.FuncAnimation

        Returns
        -------

        """
        assert (func is not None and n_frames is not None) or \
               (data is not None and data_instance is not None), \
            "Either (func, n_frames) or (data, data_instance) should be given."

        if type(axs) is Axes_plt:  # convert to array
            axs = [axs]
            data = [data]
            data_instance = [data_instance]

        assert (axs[i].figure == axs[i + 1].figure
            for i in range(len(axs) - 1)), "All axes should be of the same figure."

        if func is None:
            assert len(axs) == len(data) == len(data_instance), "Number of axes should be equal to number of data sets."
            n_frames = len(data[0])

        if titles is not None:
            kwargs = {"blit": False} | kwargs

        # if callable(titles):
        # 	titles = [titles(i) for i in range(n_frames)]  # convert to list

        def update_data(i):
            """
            Update data for animation

            Parameters
            ----------
            i : Frame index
            """
            for idx_ax, ax in enumerate(axs):
                h = data_instance[idx_ax]
                if type(h) is list:  # multiple objects
                    for j, hh in enumerate(h):
                        hh.set_data(data[idx_ax][i][j])  # update images
                else:
                    h.set_data(data[idx_ax][i])  # update image

                if titles is not None:
                    ax.set_title(titles[i])  # update title

            return data_instance

        if func is None:
            func = update_data

        self.func_animation = matplotlib.animation.FuncAnimation(fig=axs[0].figure,
                func=func, frames=n_frames,
                *args, **kwargs)
